// DSepGrid.cpp: implementation of the CDSepGrid class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "flow3d.h"
#include "DSepGrid.h"
#include "Defines.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDSepGrid::CDSepGrid()
{
	m_DSep = DEF_DSEP;
	m_InvDSep = 1.0f / DEF_DSEP;

	m_DTest = DEF_DTEST;

	m_VectorField = NULL;
	m_DSepGrid = NULL;

	m_XDim = 0;
	m_YDim = 0;
	m_ZDim = 0;

	m_QuadMinDist_start = m_DSep * m_DSep;
	m_QuadMinDist_line = m_DSep * m_DTest * m_DSep * m_DTest;

	m_MaxXIndex = 0;
	m_MaxYIndex = 0;
	m_MaxZIndex = 0;

}

CDSepGrid::CDSepGrid(CVectorField *vectorField, float dSep, float dTest)
{
	m_DSep = dSep;
	m_InvDSep = 1.0f/dSep;
	m_DTest = dTest;

	m_QuadMinDist_start = m_DSep * m_DSep;
	m_QuadMinDist_line = m_DSep * m_DTest * m_DSep * m_DTest;

	m_DSepGrid = NULL;
	
	m_XDim = 0;
	m_YDim = 0;
	m_ZDim = 0;

	m_VectorField = vectorField;

	m_MinPos = m_VectorField->m_WinMinPos;
	m_MaxPos = m_VectorField->m_WinMaxPos;

	m_MaxXIndex = 0;
	m_MaxYIndex = 0;
	m_MaxZIndex = 0;
}

CDSepGrid::~CDSepGrid()
{
	DeleteGrid();
}

void CDSepGrid::Insert(CSamplePoint *samplePoint)
{
	int x=0;
	int y=0;
	int z=0;

	GetXYZIndex(x,y,z,*samplePoint->GetPos());

	if(IsValidIndex(x,y,z))
		m_DSepGrid[x][y][z].AddHead(samplePoint);
}

void CDSepGrid::SetDSep(float dSep)
{
	m_DSep = dSep;
	m_InvDSep = 1.0f / dSep;

	m_QuadMinDist_start = m_DSep * m_DSep;
	m_QuadMinDist_line = m_DSep * m_DTest * m_DSep * m_DTest;
}

/*! 
	Purpose of dTest: Streamlines grow until they come distance 
	dTest * dSep close to other streamline.
*/
void CDSepGrid::SetDTest(float dTest)
{	
	m_DTest = dTest;
	// Recalculate minimal distance of streamline segments.
	m_QuadMinDist_line = m_DSep * m_DTest * m_DSep * m_DTest;
}

float CDSepGrid::GetDSep()
{
	return m_DSep;
}

float CDSepGrid::GetDTest()
{
	return m_DTest;
}

void CDSepGrid::SetVectorField(CVectorField *vectorField)
{
	m_VectorField = vectorField;

	m_MinPos = m_VectorField->m_WinMinPos;
	m_MaxPos = m_VectorField->m_WinMaxPos;
}

void CDSepGrid::InitGrid()
{
	//! Remove previous dsep-grid.
	DeleteGrid();
	
	//! Compute dimensions of dsep-grid.
	m_XDim = (int)(m_VectorField->GetWindowXRange()/m_DSep)+DEF_DSEPGRID_OFFS;
	m_YDim = (int)(m_VectorField->GetWindowYRange()/m_DSep)+DEF_DSEPGRID_OFFS;
	m_ZDim = (int)(m_VectorField->GetWindowZRange()/m_DSep)+DEF_DSEPGRID_OFFS;

	m_MaxXIndex = m_XDim - 1;
	m_MaxYIndex = m_YDim - 1;
	m_MaxZIndex = m_ZDim - 1;

	// Allocating the array for the dsep-grid.
	CreateGrid();
}

void CDSepGrid::DeleteGrid()
{
	if(m_DSepGrid != NULL)
	{
		for(int x=0;x<m_XDim;x++)
		{
			for(int y=0;y<m_YDim;y++)
			{
				for(int z=0;z<m_ZDim;z++)
					m_DSepGrid[x][y][z].RemoveAll();
				
				delete [] m_DSepGrid[x][y];	
			}
			delete [] m_DSepGrid[x];
		}
		delete [] m_DSepGrid;
	}
}

bool CDSepGrid::TestValid(int type, CSamplePoint *samplePoint)
{
	bool result = true;

	// Quadratic minimal distance is used instead of minimal distance, 
	// because calculation with the quadratic distance of 
	// samplepoints avoids the need of costly square root operations.
	float quadMinDist = 0.0f;

	// Select type of samplepoint: startpoint or normal streamline point?
	if (type == DEF_TYPE_START)
		quadMinDist = m_QuadMinDist_start;
	else if(type == DEF_TYPE_LINE)
		quadMinDist = m_QuadMinDist_line;

	// Calculate x,y,z-index
	int xIndex = 0;
	int yIndex = 0;
	int zIndex = 0;

	GetXYZIndex(xIndex,yIndex,zIndex,*samplePoint->GetPos());

	// Test vector length and position.
	if((samplePoint->GetMagnitude() == 0)	 ||
	   (!IsValidIndex(xIndex,yIndex,zIndex)) ||
	   (!samplePoint->GetPos()->Inside(m_MinPos,m_MaxPos)))
	{
			result = false;
	}

	// berprfung fr alle 27 mglichen Flle im 3D. Break loop if result=true.
	if(result)
	{
		for(int x = xIndex - 1; ((x < (xIndex + 2)) && result); x++)
			for(int y = yIndex - 1; ((y < (yIndex + 2)) && result); y++)
					for(int z = zIndex - 1;((z < (zIndex + 2)) && result); z++)
						if((x >= 0) && (x < m_XDim) && 
						   (y >= 0) && (y < m_YDim) &&
						   (z >= 0) && (z < m_ZDim))
						{
							CObList *samplePoints = &m_DSepGrid[x][y][z];

							// Computes quadratic mindist and returns true
							// if >= quadMinDist(=valid), else false.
							result = 
								samplePoint->ComputeQuadMinDist(samplePoints,
															    quadMinDist);
						}
	}

	return result;
}

int CDSepGrid::GetXIndex(CSamplePoint *samplePoint)
{
	return (int)((samplePoint->GetXPos() - m_MinPos.m_XPos) * m_InvDSep);
}

int CDSepGrid::GetYIndex(CSamplePoint *samplePoint)
{
	return (int)((samplePoint->GetYPos() - m_MinPos.m_YPos) * m_InvDSep);
}

int CDSepGrid::GetZIndex(CSamplePoint *samplePoint)
{
	return (int)((samplePoint->GetZPos() - m_MinPos.m_ZPos) * m_InvDSep);
}

CVectorField * CDSepGrid::GetVectorField()
{
	return m_VectorField;
}

CDSepGrid& CDSepGrid::operator=(const CDSepGrid &dSepGrid)
{
	DeleteGrid();

	m_DSep = dSepGrid.m_DSep;
	m_DTest = dSepGrid.m_DTest;
	m_InvDSep = dSepGrid.m_InvDSep;
	m_QuadMinDist_start = dSepGrid.m_QuadMinDist_start;
	m_QuadMinDist_line = dSepGrid.m_QuadMinDist_line;
	m_VectorField = dSepGrid.m_VectorField;
	
	m_XDim = dSepGrid.m_XDim;
	m_YDim = dSepGrid.m_YDim;
	m_ZDim = dSepGrid.m_ZDim;
	
	m_MinPos = dSepGrid.m_MinPos;
	m_MaxPos = dSepGrid.m_MaxPos;

	m_MaxXIndex = dSepGrid.m_MaxXIndex;
	m_MaxYIndex = dSepGrid.m_MaxYIndex;
	m_MaxZIndex = dSepGrid.m_MaxZIndex;

	CreateGrid();

	for(int x=0;x<m_XDim;x++)
		for(int y=0;y<m_YDim;y++)
			for(int z=0;z<m_ZDim;z++)
			{
				POSITION pos = dSepGrid.m_DSepGrid[x][y][z].GetHeadPosition();
				
				while(pos != NULL)
				{
					CSamplePoint *next = 
						(CSamplePoint *)dSepGrid.m_DSepGrid[x][y][z].GetNext(pos);

					m_DSepGrid[x][y][z].AddTail(next);
				}
			}

	return *this;
}

CDSepGrid::CDSepGrid(const CDSepGrid &dSepGrid)
{
	*this = dSepGrid;
}

void CDSepGrid::CreateGrid()
{
	m_DSepGrid = new CObList**[m_XDim];
	
	for(int x=0;x<m_XDim;x++)
	{
		m_DSepGrid[x] = new CObList*[m_YDim];
		
		for(int y=0;y<m_YDim;y++)
			m_DSepGrid[x][y] = new CObList[m_ZDim];
	}	
}

void CDSepGrid::GetXYZIndex(int &x, int &y, int &z, CPoint3D &pos)
{
	x = (int)((pos.m_XPos - m_MinPos.m_XPos) * m_InvDSep);
	y = (int)((pos.m_YPos - m_MinPos.m_YPos) * m_InvDSep);
	z = (int)((pos.m_ZPos - m_MinPos.m_ZPos) * m_InvDSep);
}

void CDSepGrid::Insert(CSamplePoint *samplePoint, int x, int y, int z)
{
	if(IsValidIndex(x,y,z))
		m_DSepGrid[x][y][z].AddHead(samplePoint);
}

void CDSepGrid::Clamp(int &x, int &y, int &z)
{
	if(x < 0)
		x = 0;
	else if(x >= m_XDim)
		x = m_MaxXIndex;

	if(y < 0)
		y = 0;
	else if(y >= m_YDim)
		y = m_MaxYIndex;
	
	if(z < 0)
		z = 0;
	else if(z >= m_ZDim)
		z = m_MaxZIndex;
}

bool CDSepGrid::IsValidIndex(int x, int y, int z)
{
	return !((x < 0) || (x >= m_XDim) ||
		     (y < 0) || (y >= m_YDim) ||
		     (z < 0) || (z >= m_ZDim));
}
