// Flow3DDoc.cpp : implementation of the CFlow3DDoc class
//

#include "stdafx.h"
#include "Flow3D.h"

#include "Flow3DDoc.h"

#include "EulerOp.h"
#include "RungeKutta2Op.h"
#include "RungeKutta4Op.h"

#include "NearestNeighbour.h"
#include "Trilinear.h"
#include "TrilinearVLength.h"
#include "RectTrilinear.h"
#include "RectTrilinVLength.h"

#include "StreamlineFinder.h"

#include "TestSampler.h"
#include "LorenzSampler.h"
#include "RoesslerSampler.h"

#include "MyFieldSamplingDialog.h"
#include "MyPerformanceOptDialog.h"
#include "MyProgressDialog.h"
#include "MyAddInfoDialog.h"
#include "MyResolutionsDialog.h"
#include "MyContTransfDialog.h"
#include "FLOAT.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CFlow3DDoc

IMPLEMENT_DYNCREATE(CFlow3DDoc, CDocument)

BEGIN_MESSAGE_MAP(CFlow3DDoc, CDocument)
	//{{AFX_MSG_MAP(CFlow3DDoc)
	ON_COMMAND(ID_OPTIONS_ALGORITHM, OnOptionsAlgorithm)
	ON_COMMAND(ID_OPTIONS_SETFIELDSAMPLINGOPTIONS, OnSetFieldSamplingOptions)
	ON_COMMAND(ID_OPTIONS_SETRESOLUTIONOPTIONS, OnSetResolutionOptions)
	ON_COMMAND(ID_FILE_OPENDATA, OnOpenData)
	ON_COMMAND(ID_FILE_SAVEDATAAS, OnSaveDataAs)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CFlow3DDoc construction/destruction

CFlow3DDoc::CFlow3DDoc()
{
	// TODO: add one-time construction code here
	// Set current vectorfield to NULL;
	m_CurrentField = NULL;

	// Initialise position values.
	UpdateMinMaxPos();

	// Initialise algorithm options.
	m_MaxIterations = DEF_MAXITERATIONS;
	m_IntegratorType = DEF_RK2_INDEX;
	m_InterpolatorType = DEF_TRILINEARVL_INDEX;
	m_DSepPerc = DEF_DSEP_PERC;
	m_DTest = DEF_DTEST;
	m_Dt=DEF_DT;
	m_CaPerSP = DEF_CA_PER_SP;

	// Initialise field creation options.
	m_XDim = DEF_XDIM;
	m_YDim = DEF_YDIM;
	m_ZDim = DEF_ZDIM;

	CPoint3D center = CPoint3D(DEF_XDIM/2,DEF_YDIM/2,DEF_ZDIM/2);

	m_Param1 = DEF_PARAM1;
	m_Param2 = DEF_PARAM2;
	m_Param3 = DEF_PARAM3;

	m_SamplerType = DEF_LORENZSAMPLER_INDEX;

	// Initialise statistical values.
	m_StreamlinesNum = 0;
	m_SamplePointsNum = 0;
	m_DisplSamplePNum = 0;
	m_CalculationTime = 0;

	// Initialise flags.
	m_UseDisplayLists = DEF_USEDISPLAYLISTS;
	m_IsCorrectOrder = DEF_ISCORRECTORDER;
	m_IsSampleAddInfo = DEF_ISSAMPLEADDINFO;
	m_IsTapering = DEF_ISTAPERING;
	m_IsAddInfo = DEF_ISADDINFO;
	m_IsMagicVol = DEF_ISMAGICVOLUME;
	m_IsIncremential = m_IsCalcIncr = DEF_ISINCREMENTIAL;
	m_IsCutLines = DEF_ISCUT_LINES;
	m_IsFocusContext = DEF_ISFOCUSCONTEXT;
	m_IsPlaceSeedPoint = DEF_ISPLACESEEDPOINT;
	m_IsAnimation = DEF_ISANIMATION;
	m_IsArrows = DEF_ISARROWS;

	// Initialise additional scalar info parameters.
	m_AddInfoRepr = DEF_REGIONS_INDEX;
	m_AddInfoType = DEF_VELOCITY_INDEX;

	//Init scalar regions.
	InitRegions();

	// Init minimal distance between samplepoints.
	m_MinDistPerc = DEF_MINDIST_PERC;

	// Init streamline resolution options.
	m_ResNum = m_CalcResNum = DEF_RESOLUTION_NUM;
	m_MaxResIndex = m_ResNum;

	// Set seedpoint cross appearance options.
	// Position of cross is the seedpoint of the first streamline.
	m_SeedPCross.SetLineWidth(DEF_AXIS_LINEWIDTH);
	m_SeedPCross.SetOffs(DEF_SEEDPCROSS_OFFS);

	// Create end-tapering effect table.
	m_IsEndTap = DEF_ISENDTAP;
	m_EndTapWidth = DEF_ENDTAPWIDTH;
	m_EndTapTable = CreateEndTapTable(m_EndTapWidth);
	
	// Init continuous transfer function properties.
	InitContTransfProp();

	// Set default context properties.
	m_ContextProp.resPerc = DEF_CONTEXT_RES;
	m_ContextProp.lineWidthPerc = DEF_CONTEXT_LWIDTH;
	m_ContextProp.opacityPerc = DEF_CONTEXT_OPACITY;
	m_ContextProp.material = m_RegProp[DEF_REGIONS_NUM].material;

	// Update context property absolute values.
	UpdateContextProp();

	//! Create array with minimal and maximal scalar values.
	CreateMinMaxScalars();

	// Initialise animation options.
	m_AnimSpeed = DEF_ANIMATIONTIMER_ELAPSE;
	m_AnimType = DEF_GROWING_INDEX;

	// Initialise animation index table parameters.
	m_AnimStartTbl = NULL;
	m_AnimEndTbl = NULL;
	m_AnimTblWidth = 0;
	
	// Initialise streamline recalculation type.
	m_RecalcType = DEF_NEWBOUND_INDEX;

	// Set block defaults(block dataset only).
	m_BlZOffs = 0.0f;

	GLubyte amb[4] = DEF_BLOCK_AMB;
	GLubyte dif[4] = DEF_BLOCK_DIF;
	GLubyte spec[4] = DEF_BLOCK_SPEC;

	m_BlMaterial.SetAmbientRGB(amb);
	m_BlMaterial.SetDiffuseRGB(dif);
	m_BlMaterial.SetSpecularRGB(spec);
}

CFlow3DDoc::~CFlow3DDoc()
{
	// Delete all tables.
	delete [] m_EndTapTable;
	delete [] m_MinScalars;
	delete [] m_MaxScalars;

	if(m_AnimStartTbl != NULL)
		delete [] m_AnimStartTbl;
	if(m_AnimEndTbl != NULL)
		delete [] m_AnimEndTbl;

	DeleteContents();
}

BOOL CFlow3DDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	// TODO: add reinitialization code here
	// (SDI documents will reuse this document)

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CFlow3DDoc serialization

void CFlow3DDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
		if(m_CurrentField != NULL)
		{
			int xDim = m_CurrentField->GetXDim();
			int yDim = m_CurrentField->GetYDim();
			int zDim = m_CurrentField->GetZDim();

			int size = CSamplePoint::GetAdditionalSize();
			int nT = 0;
			float dT = 0;

			// Storing header with dimension and additional information.
			char  header[DEF_HEADER_SIZE];
			sprintf(header,"SN4DB %d %d %d %d %d %f",xDim,yDim,zDim,size,nT,dT);
			ar.Write(header,DEF_BUF_SIZE);
		
			char byt[sizeof(float)];
			char *revByt;
			float posi[3];

			// Storing position information for every samplepoint.
			for(int z = 0; z < zDim; z++)
				for(int y = 0; y < yDim; y++)
					for(int x = 0; x < xDim; x++)
					{
						CSamplePoint* s = m_CurrentField->GetVField(x,y,z);
						
						posi[0] = s->GetXPos();
						posi[1] = s->GetYPos();
						posi[2] = s->GetZPos();

						for(int k=0;k<3;k++)
						{
							switch(k)
							{
								case 0:
								revByt = (char *)(&posi[0]);
								break;
								case 1:
								revByt = (char *)(&posi[1]);
								break;
								case 2:
								revByt = (char *)(&posi[2]);
								break;
							}
							
							// Reverse byte-order.
							for(int j=0;j<sizeof(float);j++)
							byt[sizeof(float)-1-j] = revByt[j];
				
							ar.Write(byt, sizeof(float));
						}
					}
		}
	}
	else
	{
		// TODO: add loading code here

		// Delete previous data.
		DeleteContents();
		
		int xDim = 0;
		int yDim = 0;
		int zDim = 0;
		int size = 0;
		int nT = 0;
		float dT = 0.0f;

		// Loadng header.
		char header[DEF_HEADER_SIZE];
		ar.Read(header,DEF_BUF_SIZE);

		sscanf(header,"SN4DB %d %d %d %d %d %f",&xDim,&yDim,&zDim,&size,&nT,&dT);
				
		CSamplePoint::SetAdditionalSize(size);
	
		m_CurrentField = new CVectorField(xDim,yDim,zDim);
		m_CurrentField->CreateField(NULL,NULL);

		CPoint3D pos;

		char byt[4];
		float posi[3];
		// loading position grid
		for(int z = 0; z < zDim;z++)
			for(int y = 0; y < yDim; y++)
			{
				for(int x = 0; x < xDim; x++)
				{
					for(int i=0;i<3;i++)
					{
						for(int j=0;j<4;j++)
							ar.Read(&byt[3-j],1);
						posi[i] = *((float *)byt);
					}
					pos.m_XPos = posi[0];
					pos.m_YPos = posi[1];
					pos.m_ZPos = posi[2];
				
					m_CurrentField->GetVField(x,y,z)->SetPos(pos);
				}
			}
		// Setting max/min-values of this field.
		m_CurrentField->InitMinMaxPos();
						
		// Update position values for document.
		UpdateMinMaxPos();
		
		// Initialise minmax-scalars array.
		delete [] m_MinScalars;
		delete [] m_MaxScalars;
	
		// Initialise document minmax-scalar array.
		CreateMinMaxScalars();

		// Initialise seedpoint cross.
		m_SeedPCross.SetPosition(CPoint3D());
		UpdateAllViews(NULL,DEF_HINT_INIT_SCENE,NULL);
  }
}

/////////////////////////////////////////////////////////////////////////////
// CFlow3DDoc diagnostics

#ifdef _DEBUG

void CFlow3DDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CFlow3DDoc commands

void CFlow3DDoc::CreateField() 
{
	// Init members.
	DeleteContents();
	m_StreamlinesNum = 0;
	m_SamplePointsNum = 0;

	// Use progress bar.
	MyProgressDialog progressDialog;

	POSITION viewpos = GetFirstViewPosition();
	CView* pFirstView = GetNextView( viewpos );

	progressDialog.Create(IDD_PROGRESSDIALOG,pFirstView);
	progressDialog.InitProgressDialog(DEF_CREATEFIELDPROGRESS_STR,
									  DEF_CREATEFIELDPROGRESS_MAX);

	progressDialog.ShowWindow(SW_SHOW);

	// Create object with calculates the sampling equations.
	CSampler *sampler = 
		MyFieldSamplingDialog::CreateSampler(m_XDim, m_YDim, m_ZDim,
											 m_Param1, m_Param2, m_Param3,
											 m_SamplerType, m_IsSampleAddInfo);


	// Set additional scalar array size.
	CSamplePoint::SetAdditionalSize(sampler->GetAdditionalSize());

	// Create field and calculate vector-values.
	m_CurrentField = new CVectorField(m_XDim,m_YDim,m_ZDim);
	m_CurrentField->CreateField(sampler,&progressDialog.m_Progress);

	// Udpate position values for vectorfield boundaries.
	UpdateMinMaxPos();
	
	// set seedpoint cross to center of field per default.
	m_SeedPCross.SetPosition(CPoint3D());

	m_AnimTblWidth = 0;

	// Delete minmax-scalar arrax
	
	delete [] m_MinScalars;
	delete [] m_MaxScalars;
	
	// Now create new minmax-scalar array and initialise with
	// default values.
	CreateMinMaxScalars();

	// Clean up.
	delete sampler;

	progressDialog.DestroyWindow();
	UpdateAllViews(NULL,DEF_HINT_INIT_SCENE,NULL);
}

void CFlow3DDoc::Recalc()
{
	// Delete all resolution streamline lists.
	DeletePiecesLists();
	DeleteStreamlines();
	
	// Seed point placing: field boundaries stay the same.
	if(m_IsPlaceSeedPoint)
	{
		// Reset seedpoint region.
		m_CurrentField->SetSeedMinMaxPos(m_MinPos,m_MaxPos);
		UpdateAllViews(NULL,DEF_HINT_INIT_SCENE,NULL);
	}
	// Focus context: focus becomes new field boundaries.
	else if(m_IsFocusContext)
	{
		// Calculate new field boundaries.
		CalcFieldMinMaxPos();
		if(m_RecalcType == DEF_NEWBOUND_INDEX)
			UpdateAllViews(NULL,DEF_HINT_INIT_SCENE,NULL);
	}
	// else: recalc original field boundaries.
	else
	{
		m_CurrentField->InitMinMaxPos();

		UpdateMinMaxPos();
		UpdateAllViews(NULL,DEF_HINT_INIT_SCENE,NULL);
	}

	// Set dsep-vectorfield to current vectorfield.
	m_DSepGrid.SetVectorField(m_CurrentField);
	
	// Disabling use of display lists during calculation of new streamlines 
	// to maintain performance.
	// Also disabling magic cube and additional info.
	bool useDisplayLists = m_UseDisplayLists;
	bool isMagicVol		 = m_IsMagicVol;
	bool isAddInfo		 = m_IsAddInfo;
	bool isFocusContext	 = m_IsFocusContext;
	bool isEndTap		 = m_IsEndTap;

	m_UseDisplayLists = false;
	m_IsAddInfo		  = false;
	m_IsMagicVol	  = false;
	m_IsFocusContext  = false;
	m_IsEndTap		  = false;

	//! Adapt resolutions to new parameter values.
	m_IsIncremential = m_IsCalcIncr;
	m_ResNum		 = m_CalcResNum;

	// Update resolution indeces.
	UpdateContextProp();
	for(int i=0;i<DEF_REGIONS_NUM + 1;i++)
		UpdateResIndex(m_RegProp[i]);
	
	// Set seedpoint from image space to proper position in vector field.
	CPoint3D seedPoint = m_SeedPCross.GetPosition();

	seedPoint.m_XPos += (m_MinPos.m_XPos + GetXRange() * 0.5f);
	seedPoint.m_YPos += (m_MinPos.m_YPos + GetYRange() * 0.5f);
	seedPoint.m_ZPos += (m_MinPos.m_ZPos + GetZRange() * 0.5f);

	// Initialise statistical values.
	m_StreamlinesNum  = 0;
	m_SamplePointsNum = 0;
	m_DisplSamplePNum = 0;

	// Take parameters from dialog and recalculate values.
	// DSep must be recalculated from percent of field x-range to absolute value.
	float dSep = GetXRange() * m_DSepPerc / DEF_PERC_100;

	// Set upper bound for streamline samplepoint calculation.
	CStreamline::SetMaxIterations(m_MaxIterations);
	
	// Create object needed for algorithm.
	CInterpolator *interpolator = 
		MyAlgoOptDialog::CreateInterpolator(m_InterpolatorType);
	
	CIntegrator *integrator = 
		MyAlgoOptDialog::CreateIntegrator(m_IntegratorType,m_CurrentField,
										  m_Dt,interpolator);

	CCandidateChooser *chooser = 
		new CCandidateChooser(&m_DSepGrid,interpolator,m_CaPerSP);
	
	CStreamlineFinder *finder = 
		new CStreamlineFinder(&m_DSepGrid,interpolator,integrator,chooser);
						
	// Time measuring variables.
	time_t start,finish;
	time(&start);

	// Use progress bar.
	MyProgressDialog progressDialog;

	POSITION viewpos  = GetFirstViewPosition();
	CView* pFirstView = GetNextView(viewpos);

	progressDialog.Create(IDD_PROGRESSDIALOG,pFirstView);
	progressDialog.InitProgressDialog(DEF_RECALCPROGRESS_STR,
									  DEF_RECALCPROGRESS_MAX);

	progressDialog.ShowWindow(SW_SHOW);

	char str[DEF_STR_SIZE];

	for(i = 1; i < m_ResNum + 1; i++)
	{
		// Index of maximal streamline resolution index.
		m_MaxResIndex = i;
			
		// String to be displayed in progress bar.
		sprintf(str,"%s (%3.1f%% of dsep)",DEF_RECALCPROGRESS_STR,
				DEF_PERC_100 * (float)i / (float)m_ResNum);

		progressDialog.SetWindowText(str);

		// Set grid-dsep to absolute dsep.
		m_DSepGrid.SetDSep(dSep * (float)m_ResNum / (float)i);
		m_DSepGrid.SetDTest(m_DTest);
		m_DSepGrid.InitGrid();
		
		// Currently processed streamline.
		CStreamline *currentLine = NULL;
	
		// Initialise streamline finder.
		finder->Init();

		// Place lower resolution streamlines into higher 
		// resolution streamline queue.
		if(m_IsIncremential)
			finder->AddStreamlines(&m_Streamlines);
		else
		// To inside test to update graphics during calculation.
			DoInsideTest();
		
		// If finder is empty=> add streamline to start algorithm.
		if(finder->GetStreamlines()->IsEmpty())
			finder->AddLine(seedPoint);
				
		// Main algorithm.
		do
		{
			// One update per algorithn step.
			progressDialog.m_Progress.StepIt();
			currentLine = finder->ApplyESS();

			if(currentLine != NULL)
			{
				//! new streamline => set resolution and add to list.
				if(currentLine->GetResolution() == -1)
				{
					currentLine->SetResolution(i);
					m_Streamlines.AddHead(currentLine);
					InsertPiecesLists(currentLine);

					// Count streamlines and samplepoints for statistics.
					m_StreamlinesNum ++;
					m_SamplePointsNum += currentLine->GetSize();
				}						
			}
			UpdateAllViews(NULL,DEF_HINT_DRAW_NEW_STREAMLINE,NULL);
		}
		while(currentLine != NULL);
	}

	//! Return members to former state.
	m_UseDisplayLists = useDisplayLists;
	m_IsAddInfo = isAddInfo;
	m_IsMagicVol = isMagicVol;
	m_IsFocusContext = isFocusContext;
	m_IsEndTap = isEndTap;

	// Destroy progress dialog.
	progressDialog.DestroyWindow();

	// Calculate minimal and maximal scalar values
	// of newly calculated streamlines.
	CalcMinMaxScalars();

	//! Take calculation time for statistics.
	time(&finish);
	m_CalculationTime = (long)difftime(finish,start);

	// Create table used to hold indexes for animation.
	if(m_AnimStartTbl != NULL)
		delete [] m_AnimStartTbl;
	if(m_AnimEndTbl != NULL)
		delete [] m_AnimEndTbl;

	CreateAnimIndexTbl();

	// Update view for the last time to display all changes.
	UpdateAllViews(NULL,DEF_HINT_REDRAW_SCENE,NULL);

	// Now clean up.
	delete interpolator;
	delete integrator;
	delete chooser;
	delete finder;
}

void CFlow3DDoc::UpdateMinMaxPos()
{
	if(m_CurrentField != NULL)
	{
		m_MinPos = m_CurrentField->m_WinMinPos;
		m_MaxPos = m_CurrentField->m_WinMaxPos;
	}
	else
	{
		m_MinPos = CPoint3D();
		m_MaxPos = CPoint3D();
	}
}

void CFlow3DDoc::DeleteContents() 
{
	// TODO: Add your specialized code here and/or call the base class
	if(m_CurrentField != NULL)
	{
		delete m_CurrentField;
		m_CurrentField = NULL;
	}

	DeletePiecesLists();
	DeleteStreamlines();

	CDocument::DeleteContents();
}

void CFlow3DDoc::OnOptionsAlgorithm() 
{
	// TODO: Add your command handler code here

	MyAlgoOptDialog algoOptDialog;

	// Initialise algorithm options dialog values.
	algoOptDialog.InitAlgoOptDlg(m_DSepPerc,m_DTest,m_Dt,m_CaPerSP,
								 m_MaxIterations,m_IntegratorType,m_InterpolatorType);

	if(algoOptDialog.DoModal() == IDOK)
	{
		m_DSepPerc = algoOptDialog.GetDSepPerc();
		m_DTest = algoOptDialog.GetDTest();
		m_Dt = algoOptDialog.GetDt();

		m_MaxIterations = algoOptDialog.GetMaxIterations();
		m_CaPerSP = algoOptDialog.GetCaPerSP();
			
		m_IntegratorType = algoOptDialog.GetIntegratorType();
		m_InterpolatorType = algoOptDialog.GetInterpolatorType();
	}
}

void CFlow3DDoc::OnSetFieldSamplingOptions()
{
	// TODO: Add your command handler code here
	MyFieldSamplingDialog fieldSamplingDialog;

	fieldSamplingDialog.SetParam1(m_Param1);
	fieldSamplingDialog.SetParam2(m_Param2);
	fieldSamplingDialog.SetParam3(m_Param3);

	fieldSamplingDialog.SetXDim(m_XDim);
	fieldSamplingDialog.SetYDim(m_YDim);
	fieldSamplingDialog.SetZDim(m_ZDim);

	fieldSamplingDialog.SetSamplerType(m_SamplerType);
	fieldSamplingDialog.SetIsSampleAddInfo(m_IsSampleAddInfo);
	if(fieldSamplingDialog.DoModal()==IDOK)
	{
		m_XDim = fieldSamplingDialog.GetXDim();
		m_YDim = fieldSamplingDialog.GetYDim();
		m_ZDim = fieldSamplingDialog.GetZDim();

		m_Param1 = fieldSamplingDialog.GetParam1();
		m_Param2 = fieldSamplingDialog.GetParam2();
		m_Param3= fieldSamplingDialog.GetParam3();

		m_SamplerType = fieldSamplingDialog.GetSamplerType();
		m_IsSampleAddInfo = fieldSamplingDialog.GetIsSampleAddInfo();
	}
}

void CFlow3DDoc::InsertPiecesLists(CStreamline *streamline)
{
	POSITION pos = streamline->GetSamplePoints()->GetHeadPosition();

	float vals[DEF_CONTTRANSFVALUES_NUM];
	GetMinMaxScalar(m_AddInfoType, vals[DEF_MINPOS_INDEX],vals[DEF_MAXPOS_INDEX]);

	vals[DEF_INTERPOS_INDEX] = vals[DEF_MINPOS_INDEX] + (vals[DEF_MAXPOS_INDEX] - 
							   vals[DEF_MINPOS_INDEX]) * m_InterPos;

	CSamplePoint *last = NULL;
	
	if(pos != NULL)
		last = (CSamplePoint *)streamline->GetSamplePoints()->GetNext(pos);

	// Initialise with random index.
	//int i = rand();
	int i = 0;

	bool isStart = true;

	POSITION xPos = m_XList.GetHeadPosition();
	POSITION yPos = m_YList.GetHeadPosition();
	POSITION zPos = m_ZList.GetHeadPosition();

	CObList *list = new CObList();
	
	m_LinePieces.AddHead((CObject *)list);

	// Quadratic absolute minimal distance between samplepoints.
	float quadMinDist = GetXRange() * m_MinDistPerc / DEF_PERC_100;
	quadMinDist *= quadMinDist;

	while(pos != NULL)
	{
		CSamplePoint *next;
		
		do
		{
			next = (CSamplePoint *)streamline->GetSamplePoints()->GetNext(pos);
		}
		while( (pos != NULL) && (last->GetQuadDistance(next) < quadMinDist));
		
		// distance >= quadMinDist => insert piece
		if(last->GetQuadDistance(next) >= quadMinDist)
		{
			CStreamlinePiece *piece = new CStreamlinePiece(last, next, i, streamline); 

			piece->m_IsLineStart = isStart;
			// Increase pieces count by one.
			m_DisplSamplePNum ++;

			//! Update piece values.
			UpdatePiece(piece, vals);

			list->AddHead(piece);

			xPos = InsertXList(piece, xPos, m_IsCorrectOrder);
			
			// If correct renderng order, then also insert into y and z-list.
			// Else only x-list is used to display streamlines.
			if(m_IsCorrectOrder)
			{
				yPos = InsertYList(piece, yPos, m_IsCorrectOrder);
				zPos = InsertZList(piece, zPos, m_IsCorrectOrder);
			}

			last = next;
			if(piece->m_Index >= INT_MAX)
				piece->m_Index = 0;

			i++;
			isStart = false;
		}
	}
}

POSITION CFlow3DDoc::InsertXList(CStreamlinePiece *piece, POSITION pos, bool sort)
{
	POSITION result = NULL;

		if((pos != NULL)&&(sort))
	{
		CStreamlinePiece *listPiece = (CStreamlinePiece *)m_XList.GetAt(pos);
		POSITION lastPos = pos;

		if(listPiece->GetStartPXPos() < piece->GetStartPXPos())
		{
			while((pos != NULL) && (listPiece->GetStartPXPos() < piece->GetStartPXPos()))
			{
				lastPos = pos;
				listPiece = (CStreamlinePiece *)m_XList.GetNext(pos);
			}
			result = m_XList.InsertAfter(lastPos,piece);
		}
		else
		{
			while((pos != NULL) && (listPiece->GetStartPXPos() > piece->GetStartPXPos()))
			{
				lastPos = pos;
				listPiece = (CStreamlinePiece *)m_XList.GetPrev(pos);
			}
			result = m_XList.InsertBefore(lastPos,piece);
		}
	}
	else 
	{
		result = m_XList.AddTail(piece);
	}

	return result;
}

POSITION CFlow3DDoc::InsertYList(CStreamlinePiece *piece, POSITION pos, bool sort)
{
	POSITION result = NULL;

	if((pos != NULL)&&(sort))
	{
		CStreamlinePiece *listPiece = (CStreamlinePiece *)m_YList.GetAt(pos);
		POSITION lastPos = pos;

		if(listPiece->GetStartPYPos() < piece->GetStartPYPos())
		{
			while((pos != NULL) && (listPiece->GetStartPYPos() < piece->GetStartPYPos()))
			{
				lastPos = pos;
				listPiece = (CStreamlinePiece *)m_YList.GetNext(pos);
			}
			result = m_YList.InsertAfter(lastPos,piece);
		}
		else
		{
			while((pos != NULL) && (listPiece->GetStartPYPos() > piece->GetStartPYPos()))
			{
				lastPos = pos;
				listPiece = (CStreamlinePiece *)m_YList.GetPrev(pos);
			}
			result = m_YList.InsertBefore(lastPos,piece);
		}
	}
	else 
	{
		result = m_YList.AddTail(piece);
	}

	return result;
}

POSITION CFlow3DDoc::InsertZList(CStreamlinePiece *piece, POSITION pos, bool sort)
{
	POSITION result = NULL;

	if((pos != NULL)&&(sort))
	{
		CStreamlinePiece *listPiece = (CStreamlinePiece *)m_ZList.GetAt(pos);
		POSITION lastPos = pos;

		if(listPiece->GetStartPZPos() < piece->GetStartPZPos())
		{
			while((pos != NULL) && (listPiece->GetStartPZPos() < piece->GetStartPZPos()))
			{
				lastPos = pos;
				listPiece = (CStreamlinePiece *)m_ZList.GetNext(pos);
			}
			result = m_ZList.InsertAfter(lastPos,piece);
		}
		else
		{
			while((pos != NULL) && (listPiece->GetStartPZPos() > piece->GetStartPZPos()))
			{
				lastPos = pos;
				listPiece = (CStreamlinePiece *)m_ZList.GetPrev(pos);
			}
			result = m_ZList.InsertBefore(lastPos,piece);
		}
	}
	else 
	{
		result = m_ZList.AddTail(piece);
	}

	return result;
}

CObList * CFlow3DDoc::GetXList()
{
	return &m_XList;
}

CObList * CFlow3DDoc::GetYList()
{
	return &m_YList;
}

CObList * CFlow3DDoc::GetZList()
{
	return &m_ZList;
}

void CFlow3DDoc::DeletePiecesLists()
{
	m_YList.RemoveAll();
	m_ZList.RemoveAll();
	m_XList.RemoveAll();

	POSITION pos = m_LinePieces.GetHeadPosition();
	
	// Just delete objects of main pieces list.
	// because all lists contain pointer to same pieces-objects.
	while(pos != NULL)
	{
		CObList *list  = (CObList *)m_LinePieces.GetNext(pos);

		POSITION listPos = list->GetHeadPosition();

		while(listPos != NULL)
		{
			CStreamlinePiece *piece  = (CStreamlinePiece *)list->GetNext(listPos);
			list->RemoveHead();
			delete piece;
		}

		m_LinePieces.RemoveHead();
		delete list;
	}
}

void CFlow3DDoc::CalcMinMaxScalars()
{
	for(int i = 0; i < DEF_SCALAR_OVERALL; i++)
	{
		m_MinScalars[i] = FLT_MAX;
		m_MaxScalars[i] = FLT_MIN;
	}

	// Set orientation minmax-boundaries beforehand.
	m_MinScalars[DEF_ORIENT_X_INDEX] = 0.0f;
	m_MinScalars[DEF_ORIENT_Y_INDEX] = 0.0f;
	m_MinScalars[DEF_ORIENT_Z_INDEX] = 0.0f;

	m_MaxScalars[DEF_ORIENT_X_INDEX] = PI;
	m_MaxScalars[DEF_ORIENT_Y_INDEX] = PI;
	m_MaxScalars[DEF_ORIENT_Z_INDEX] = PI;

	POSITION pos = m_XList.GetHeadPosition();

	// Except orientation, set boundaries by getting minmax-values
	// of streamline pieces.
	while(pos != NULL)
	{
		CStreamlinePiece *piece = (CStreamlinePiece *)m_XList.GetNext(pos);
			
		for(i = 0; i < DEF_SCALAR_OVERALL; i++)
		{	
			float scalar = GetScalar(i,piece);

			if(scalar < m_MinScalars[i])
				m_MinScalars[i] = scalar;
			if(scalar > m_MaxScalars[i])
				m_MaxScalars[i] = scalar;
		}
	}
}

void CFlow3DDoc::ResortPiecesLists()
{
	// Delete former lists.
	DeletePiecesLists();
	
	// Initialise pieces count.
	m_DisplSamplePNum = 0;

	// Use progress bar because sorting the lists could be a long task. 
	MyProgressDialog progressDialog;

	POSITION viewpos = GetFirstViewPosition();
	CView* pFirstView = GetNextView( viewpos );

	progressDialog.Create(IDD_PROGRESSDIALOG,pFirstView);
	progressDialog.InitProgressDialog(DEF_UPDATEPLISTSPROGRESS_STR,
									  DEF_UPDATEPLISTSPROGRESS_MAX);
	progressDialog.ShowWindow(SW_SHOW);

	POSITION pos = m_Streamlines.GetHeadPosition();

	while(pos != NULL)
	{
		progressDialog.m_Progress.StepIt();

		CStreamline *next = (CStreamline *)m_Streamlines.GetNext(pos);
		InsertPiecesLists(next);
	}
	
	// As animation index tables hold pieces-lists list positions, resorting the
	// lists also means that the animation index tables have to be recalculated.
	if(m_AnimStartTbl != NULL)
		delete [] m_AnimStartTbl;
	if(m_AnimEndTbl != NULL)
		delete [] m_AnimEndTbl;

	CreateAnimIndexTbl();

	// Do inside test to determine which streamlines to render.
	DoInsideTest();
	
	progressDialog.DestroyWindow();
}

CVectorField *CFlow3DDoc::GetCurrentField()
{
	return m_CurrentField;
}

int CFlow3DDoc::GetResNum()
{
	return m_ResNum;
}

float CFlow3DDoc::GetMaxXPos()
{
	return m_MaxPos.m_XPos;
}

float CFlow3DDoc::GetMaxYPos()
{
	return m_MaxPos.m_YPos;
}

float CFlow3DDoc::GetMaxZPos()
{
	return m_MaxPos.m_ZPos;
}

float CFlow3DDoc::GetMinXPos()
{
	return m_MinPos.m_XPos;
}

float CFlow3DDoc::GetMinYPos()
{
	return m_MinPos.m_YPos;
}

float CFlow3DDoc::GetMinZPos()
{
	return m_MinPos.m_ZPos;
}

float CFlow3DDoc::GetXRange()
{
	return m_MaxPos.m_XPos - m_MinPos.m_XPos;
}

float CFlow3DDoc::GetYRange()
{
	return m_MaxPos.m_YPos - m_MinPos.m_YPos;
}

float CFlow3DDoc::GetZRange()
{
	return m_MaxPos.m_ZPos - m_MinPos.m_ZPos;
}

bool CFlow3DDoc::GetUseDisplayLists()
{
	return m_UseDisplayLists;
}

bool CFlow3DDoc::GetIsMagicVol()
{
	return m_IsMagicVol;
}

int CFlow3DDoc::GetStreamlinesNum()
{
	return m_StreamlinesNum;
}

long CFlow3DDoc::GetSamplePointsNum()
{
	return m_SamplePointsNum;
}

long CFlow3DDoc::GetDisplSamplePNum()
{
	return m_DisplSamplePNum;
}

long CFlow3DDoc::GetCalculationTime()
{
	return m_CalculationTime;
}

bool CFlow3DDoc::GetIsCorrectOrder()
{
	return m_IsCorrectOrder;
}

bool CFlow3DDoc::GetIsTapering()
{
	return m_IsTapering;
}

bool CFlow3DDoc::GetIsAddInfo()
{
	return m_IsAddInfo;
}

bool CFlow3DDoc::GetIsCutLines()
{
	return m_IsCutLines;
}


float CFlow3DDoc::GetInterPos()
{
	return m_InterPos;
}

float CFlow3DDoc::GetMinDistPerc()
{
	return m_MinDistPerc;
}

bool CFlow3DDoc::GetIsIncremential()
{
	return m_IsIncremential;
}

bool CFlow3DDoc::GetIsCalcIncr()
{
	return m_IsCalcIncr;
}

int CFlow3DDoc::GetAddInfoType()
{
	return m_AddInfoType;
}

float CFlow3DDoc::GetDSepPerc()
{
	return m_DSepPerc;
}

float CFlow3DDoc::GetDTest()
{
	return m_DTest;
}

float CFlow3DDoc::GetDt()
{
	return m_Dt;
}

int CFlow3DDoc::GetMaxIterations()
{
	return m_MaxIterations;
}

int CFlow3DDoc::GetCaPerSP()
{
	return m_CaPerSP;
}

bool CFlow3DDoc::GetIsFocusContext()
{
	return m_IsFocusContext;
}

MyAxisCross *CFlow3DDoc::GetSeedPCross()
{
	return &m_SeedPCross;
}

void CFlow3DDoc::SetMinDistPerc(float minDistPerc)
{
	m_MinDistPerc = minDistPerc;
}

void CFlow3DDoc::SetIsFocusContext(bool isFocusContext)
{
	m_IsFocusContext = isFocusContext;
}

void CFlow3DDoc::SetAddInfoType(int addInfoType)
{
	m_AddInfoType = addInfoType;
}

void CFlow3DDoc::SetIsCorrectOrder(bool isCorrectOrder)
{
	m_IsCorrectOrder = isCorrectOrder;
}

void CFlow3DDoc::SetIsCutLines(bool isCutLines)
{
	m_IsCutLines = isCutLines;
}

void CFlow3DDoc::SetIsAddInfo(bool isAddInfo)
{
	m_IsAddInfo = isAddInfo;
}

void CFlow3DDoc::SetInterPos(float interPos)
{
	m_InterPos = interPos;
}

void CFlow3DDoc::SetIsMagicVol(bool isMagicVol)
{
	m_IsMagicVol = isMagicVol;
}

void CFlow3DDoc::SetIsTapering(bool isTapering)
{
	m_IsTapering = isTapering;
}

void CFlow3DDoc::SetUseDisplayLists(bool useDisplayLists)
{
	m_UseDisplayLists = useDisplayLists;
}

void CFlow3DDoc::UpdatePiecesLists()
{
	POSITION pos = m_LinePieces.GetHeadPosition();

	float vals[DEF_CONTTRANSFVALUES_NUM];

	GetMinMaxScalar(m_AddInfoType, vals[DEF_MINPOS_INDEX],vals[DEF_MAXPOS_INDEX]);
	
	vals[DEF_INTERPOS_INDEX] = vals[DEF_MINPOS_INDEX] + 
		(vals[DEF_MAXPOS_INDEX] - vals[DEF_MINPOS_INDEX]) * m_InterPos;

	while(pos != NULL)
	{
		CObList * list = (CObList *)m_LinePieces.GetNext(pos);

		POSITION linePos = list->GetHeadPosition();
		
		while(linePos != NULL)
		{
			CStreamlinePiece *piece = (CStreamlinePiece *)list->GetNext(linePos);
			UpdatePiece(piece,vals);
		}
	}
	// If necessary, do inside-test to determine which streamlines to render,
	// which streamlines belong to focus and which to context.
	if(m_IsFocusContext || m_IsAddInfo || m_IsEndTap || 
	   m_IsAnimation	|| m_IsArrows)
		DoInsideTest();
}

void CFlow3DDoc::UpdatePiece(CStreamlinePiece *piece, float *vals)
{
	//float lineWidth = m_LineWidth;
	float lineWidth = m_RegProp[DEF_REGIONS_NUM].lineWidth;
	int resIndex = m_ResNum;

	GLubyte baseColor[4];
	
	for(int i=0;i<3;i++)
		baseColor[i] = DEF_COLOR_MAX;

	baseColor[3] = DEF_OPACITY_MAX;
	
	// Set to default regions index.
	byte regIndex = DEF_REGIONS_NUM;

	// Map vals-array to scalar min-max-interval.
	vals[DEF_MINPOS_INDEX] = m_MinScalars[m_AddInfoType];
	vals[DEF_MAXPOS_INDEX] = m_MaxScalars[m_AddInfoType];

	// Properties of continuous transfer function.
	ContTransfProp currProp;

	if(m_IsAddInfo)
	{
		float scalar = GetScalar(m_AddInfoType,piece);

		if(m_AddInfoRepr == DEF_REGIONS_INDEX)
		{
				int index = 
					MyEditRegionsDialog::
						MapValue2PropNr(m_RegProp, DEF_REGIONS_NUM, 
						  			    scalar, vals[DEF_MINPOS_INDEX], 
										vals[DEF_MAXPOS_INDEX]);

				regIndex = index;
				resIndex = m_RegProp[index].resIndex;
				lineWidth = m_RegProp[index].lineWidth;
		}
		else
		{
			MyContTransfDialog::MapValue2Prop_3(currProp,scalar,
												m_ContTransfProp,vals);
			
			for(i=0;i < 4;i++)
				baseColor[i] = currProp.rgba[i];

			// Calculate resolution index from percentage.
			resIndex = currProp.res * m_ResNum / 100;
			lineWidth = currProp.lineWidth;
		}
	}

	if(m_IsTapering)
		lineWidth *= ThicknessCoeff(piece->m_StartPoint);

	for(i=0;i<4;i++)
		piece->m_Color[i] = baseColor[i];

	piece->m_LineWidth = lineWidth;
	piece->m_RegIndex = regIndex;
	piece->m_IsInside = true;
	piece->m_IsDraw = true;
	piece->m_IsLineStart = false;
	piece->m_ResIndex = resIndex;
	piece->m_EndTapIndex = 0;
}

float CFlow3DDoc::ThicknessCoeff(CSamplePoint *samplePoint)
{

	float result = 1.0f;
	
	float dSep = m_DSepGrid.GetDSep();
	float absDTest = dSep * m_DSepGrid.GetDTest();

	float minDist = sqrt(samplePoint->GetQuadMinDist());

	if(minDist >= dSep)
	{
		result = 1.0f;
	}
	else if((dSep - absDTest) > 0)
	{
		result = (minDist - absDTest)/(dSep - absDTest);
	}
	
	return result;
}

float CFlow3DDoc::GetScalar(int index, CStreamlinePiece *piece)
{
	CVector3D *tang = piece->m_StartPoint->GetUnitVec();
	CSamplePoint *s = piece->m_StartPoint;
	float result=0;

	switch( index ) 
	{
		case DEF_VELOCITY_INDEX: 
			result = s->GetMagnitude();
			break;
		case DEF_ORIENT_X_INDEX: 
			result = acos(tang->m_XComp);
			break;
		case DEF_ORIENT_Y_INDEX: 
			result = acos(tang->m_YComp);
			break;
		case DEF_ORIENT_Z_INDEX: 
			result = acos(tang->m_ZComp);
			break;
		case DEF_POS_X_INDEX:
			result = s->GetXPos();
			break;
		case DEF_POS_Y_INDEX:
			result = s->GetYPos();
			break;
		case DEF_POS_Z_INDEX:
			result = s->GetZPos();
			break;
		case DEF_DIF_ORIENT_INDEX:
			result = piece->CalcDifOrient();
			break;
		case DEF_DIF_VLEN_INDEX:
			result = piece->CalcDifVLen();
			break;
		default: 
			// Any other scalar info given in the vectorfield.
			int addIndex = index - DEF_SCALAR_NUM;
			result = s->GetAdditional(addIndex);
			break;
	}
	return result;
}

void CFlow3DDoc::DeleteStreamlines()
{
	POSITION pos = m_Streamlines.GetHeadPosition();

	while(pos != NULL)
	{
		CStreamline * s = (CStreamline *)m_Streamlines.GetNext(pos);
		m_Streamlines.RemoveHead();
		delete s;
	}
}

int CFlow3DDoc::GetInterpolatorType()
{
	return m_InterpolatorType;
}

int CFlow3DDoc::GetIntegratorType()
{
	return m_IntegratorType;
}

void CFlow3DDoc::OnSetResolutionOptions() 
{
	// TODO: Add your command handler code here
	MyResolutionsDialog resolutionsDialog;

	resolutionsDialog.SetIsIncremential(m_IsCalcIncr);
	resolutionsDialog.SetResNum(m_CalcResNum);

	if(resolutionsDialog.DoModal() == IDOK)
	{
		m_IsCalcIncr = resolutionsDialog.GetIsIncremential();
		m_CalcResNum = resolutionsDialog.GetResNum();		
	}	
}

void CFlow3DDoc::UpdateResIndex(RegProp &prop)
{
	prop.resIndex = prop.res * m_ResNum / 100;
}

int CFlow3DDoc::GetEndTapWidth()
{
	return m_EndTapWidth;
}

void CFlow3DDoc::SetEndTapWidth(int width)
{
	m_EndTapWidth = width;
}

int CFlow3DDoc::CalcEndTapIndex(int index, int width, int lineLength)
{
	if(index >= (lineLength >> 1))
		index = lineLength - index - 1;

	if(index >= width)
		index = width - 1;

	return index;
}

void CFlow3DDoc::SetIsEndTap(bool isEndTap)
{
	m_IsEndTap = isEndTap;
}

bool CFlow3DDoc::GetIsEndTap()
{
	return m_IsEndTap;
}

void CFlow3DDoc::CalcArrow_EndTap()
{
	POSITION pos = m_LinePieces.GetHeadPosition();
	while(pos != NULL)
	{
		CObList * list = (CObList *)m_LinePieces.GetNext(pos);

		POSITION linePos = list->GetHeadPosition();
		
		bool isLine = false;
				
		int i=0;

		POSITION firstPos = linePos;
		POSITION prevPos = linePos;
		
		int res = 0;
		
		if(linePos != NULL)
			res = ((CStreamlinePiece *)list->GetAt(linePos))->
				GetStreamline()->GetResolution();

		while(linePos != NULL)
		{
			prevPos = linePos;
			CStreamlinePiece *piece = (CStreamlinePiece *)list->GetNext(linePos);
						
			bool isDraw = piece->m_IsDraw;

			if((!isLine) && isDraw)
			{
				isLine = true;
				firstPos = linePos;
				i = 0;
				piece->m_EndTapIndex = 0;
			}
			else if(isLine && (!isDraw) || (linePos == NULL))
			{

				isLine = false;
				int j=1;
				
				if(isDraw)
				{
					prevPos = linePos;
					i ++;
				}

				while(firstPos != prevPos)
				{
					piece = (CStreamlinePiece *)list->GetNext(firstPos);
					piece->m_EndTapIndex = CalcEndTapIndex(j,m_EndTapWidth,i);
					j++;
					if(firstPos == prevPos)
						piece->m_IsLineStart = true;
				}
					
			}			
			if(isLine)
				i++;
		}
	}
}

float * CFlow3DDoc::CreateEndTapTable(int width)
{
	float *result = new float[width];

	float widthf = (float)width;
	for(int i = 0; i < width; i++)
		result[i] = (float)(i + 1)/widthf;

	return result;
}

bool CFlow3DDoc::IsDraw(bool &isInside, CStreamlinePiece *piece, int res, 
						int focusRes, int contextRes)
{
	bool isDraw = false;

	if(m_IsIncremential || (res == focusRes))
		isDraw = true;

	if(m_IsFocusContext)
	{
		isDraw = false;

		isInside = m_MagicVolume->IsInside(*piece->m_StartPoint->GetPos());
						
		if(m_IsIncremential)
		{
			if(isInside || (res <= contextRes))
				isDraw = true;
		}
		else
		{
			if(isInside)
			{
				if((res == focusRes) || ((!m_IsCutLines) && (res == contextRes)))
					isDraw = true;
				
			}
			else if(res == contextRes)
			{
				isDraw = true;
			}
		}
	}
	else if(m_IsAddInfo)
	{
		isDraw = false;
		if(m_IsIncremential)
		{
			if(piece->m_ResIndex >= res)
				isDraw = true;
		}
		else
		{
			if(piece->m_ResIndex == res)
				isDraw = true;
		}
	}

	return isDraw;
}

void CFlow3DDoc::GetMinMaxScalar(int index, float &minScalar, float &maxScalar)
{
	minScalar = m_MinScalars[index];
	maxScalar = m_MaxScalars[index];
}

void CFlow3DDoc::CalcIsDraw()
{
	POSITION pos = m_LinePieces.GetHeadPosition();
	UpdateContextProp();
	
	while(pos != NULL)
	{
		int j=0;

		CObList * list = (CObList *)m_LinePieces.GetNext(pos);

		POSITION linePos = list->GetHeadPosition();
		
		int contextRes = m_ContextProp.resIndex;
		int res = 0;
		
		if(linePos != NULL)
			res = ((CStreamlinePiece *)list->GetAt(linePos))->
				GetStreamline()->GetResolution();

		while(linePos != NULL)
		{
			CStreamlinePiece *piece = (CStreamlinePiece *)list->GetNext(linePos);

			piece->m_IsDraw = 
					IsDraw(piece->m_IsInside,piece,res,
						   m_MaxResIndex,contextRes);
			
			piece->m_IsLineStart = false;		
		}			
				
	}
}

void CFlow3DDoc::SetContTransfProp(int index, ContTransfProp &prop)
{
	m_ContTransfProp[index].lineWidth = prop.lineWidth;
	m_ContTransfProp[index].res = prop.res;
	
	for(int i = 0; i < 4; i++)
		m_ContTransfProp[index].rgba[i] = prop.rgba[i];
}

void CFlow3DDoc::GetContTransfProp(int index, ContTransfProp &prop)
{
	prop.lineWidth = m_ContTransfProp[index].lineWidth;
	prop.res = m_ContTransfProp[index].res;

	for(int i = 0; i < 4; i++)
		prop.rgba[i] = m_ContTransfProp[index].rgba[i];
}

void CFlow3DDoc::CreateMinMaxScalars()
{
	m_MinScalars = new float[DEF_SCALAR_OVERALL];
	m_MaxScalars = new float[DEF_SCALAR_OVERALL];
	CalcMinMaxScalars();
}

void CFlow3DDoc::SetAddInfoRepr(int repr)
{
	m_AddInfoRepr = repr;
}

int CFlow3DDoc::GetAddInfoRepr()
{
	return m_AddInfoRepr;
}

void CFlow3DDoc::UpdateContextProp()
{
	m_ContextProp.lineWidth = 
		(int)(m_RegProp[DEF_REGIONS_NUM].lineWidth * 
			  m_ContextProp.lineWidthPerc / DEF_PERC_100);
	if(m_ContextProp.lineWidth < DEF_LINEWIDTH_MIN)
		m_ContextProp.lineWidth = DEF_LINEWIDTH_MIN;

	m_ContextProp.resIndex = 
		m_ResNum * m_ContextProp.resPerc / (int)DEF_PERC_100;

	m_ContextProp.opacity = 
		(GLubyte)((m_RegProp[DEF_REGIONS_NUM].material.m_Opacity * 
				   m_ContextProp.opacityPerc ) / DEF_PERC_100);
}

void CFlow3DDoc::DoInsideTest()
{
	
	float xOffs = GetXRange() * 0.5f + GetMinXPos();
	float yOffs = GetYRange() * 0.5f + GetMinYPos();
	float zOffs = GetZRange() * 0.5f + GetMinZPos();

	// Translate magic volume for isinside-test.
	CPoint3D oldPos = m_MagicVolume->GetPosition();
	CPoint3D newPos = CPoint3D(oldPos.GetXPos()+xOffs,
							   oldPos.GetYPos()+yOffs,
							   oldPos.GetZPos()+zOffs);

			
	m_MagicVolume->SetPosition(newPos);

	if((!m_IsIncremential) || m_IsAddInfo || m_IsFocusContext || 
	   m_IsAnimation)
	{
		CalcIsDraw();

		if(m_IsAnimation)
			UpdateAnimIndex();
	}

	if(m_IsEndTap || m_IsArrows)
		CalcArrow_EndTap();
	
	m_MagicVolume->SetPosition(oldPos);
}

void CFlow3DDoc::CreateAnimIndexTbl()
{
	m_AnimTblWidth = m_StreamlinesNum;
	m_AnimStartTbl = new POSITION[m_AnimTblWidth];
	m_AnimEndTbl = new POSITION[m_AnimTblWidth];

	POSITION pos = m_LinePieces.GetHeadPosition();

	int i=0;
	while((i<m_AnimTblWidth) && (pos != NULL))
	{
		CObList * list = (CObList *)m_LinePieces.GetNext(pos);
		
		POSITION linePos = list->GetHeadPosition();
		m_AnimStartTbl[i] = linePos;
		m_AnimEndTbl[i] = linePos;
		i++;
	}
		
}

void CFlow3DDoc::UpdateAnimIndex()
{
	POSITION pos = m_LinePieces.GetHeadPosition();

	int i = 0;
	
	if(m_AnimType == DEF_GROWING_INDEX)
	{
		while(pos != NULL)
		{
			CObList * list = (CObList *)m_LinePieces.GetNext(pos);

			POSITION currentPos = list->GetHeadPosition();

			if(m_AnimStartTbl[i] == NULL)
			{
				m_AnimStartTbl[i] = list->GetHeadPosition();
				m_AnimEndTbl[i] = list->GetHeadPosition();
			}

			while(currentPos != m_AnimStartTbl[i])
			{
				CStreamlinePiece *piece = (CStreamlinePiece *)list->GetNext(currentPos);
				if(piece->m_IsInside)
					piece->m_IsDraw = false;
			}

			currentPos = m_AnimEndTbl[i];

			while(currentPos != NULL)
			{
				CStreamlinePiece *piece = (CStreamlinePiece *)list->GetNext(currentPos);
				if(piece->m_IsInside)
					piece->m_IsDraw = false;
			}

			if(m_AnimEndTbl[i] == NULL)
			{
				if(m_AnimStartTbl[i] != NULL)
					(void)list->GetNext(m_AnimStartTbl[i]);
			}
			else 
			{
				(void)list->GetNext(m_AnimEndTbl[i]);
			}
			i++;
		}
	}
	else
	{
		while(pos != NULL)
		{
			CObList * list = (CObList *)m_LinePieces.GetNext(pos);

			POSITION linePos = list->GetHeadPosition();
			int size = 0;

			if(linePos != NULL)
				size = ((CStreamlinePiece *)list->GetAt(linePos))->
					m_Streamline->m_Size;

			while(linePos != NULL)
			{
				CStreamlinePiece *piece = (CStreamlinePiece *)list->GetNext(linePos);
			
				// animation => move index
				if(piece->m_Index >= INT_MAX )
					piece->m_Index = 0;
				piece->m_Index ++;
							
			}
		}
	}
}

/*!  If field is recalculated using focus/context mode, there are two options:
	 1) Recalculate field with focus as new field boundaries.
	 3) Recalculate field with focus as seedpoint region(the region where new
		streamlines can be planted.

	 This calculates caculates the field parameters according to one of the two 
	 upper options.
*/
void CFlow3DDoc::CalcFieldMinMaxPos()
{
	CPoint3D minPoint0;
	CPoint3D maxPoint0;
			
	if(m_MagicVolType == DEF_MCUBE_INDEX)
	{
		MyCube *cube = (MyCube *)m_MagicVolume;

		minPoint0 = cube->GetMinPoint();
		maxPoint0 = cube->GetMaxPoint();
	}
	else if(m_MagicVolType == DEF_MRECT_INDEX)
	{
		MyRectangle *rect = (MyRectangle *)m_MagicVolume;

		minPoint0 = rect->GetMinPoint();
		maxPoint0 = rect->GetMaxPoint();
	}
	else if(m_MagicVolType == DEF_MSPHERE_INDEX)
	{
		MySphere *sphere = (MySphere *)m_MagicVolume;

		CPoint3D pos = sphere->GetPosition();

		minPoint0 = CPoint3D(pos.m_XPos - sphere->GetRadius(),
							 pos.m_YPos - sphere->GetRadius(),
							 pos.m_ZPos - sphere->GetRadius());

		maxPoint0 = CPoint3D(pos.m_XPos + sphere->GetRadius(),
							 pos.m_YPos + sphere->GetRadius(),
							 pos.m_ZPos + sphere->GetRadius());
	}

	CPoint3D minPoint1 = m_Boundaries.GetMinPoint();
	CPoint3D maxPoint1 = m_Boundaries.GetMaxPoint();

	float xOffs = GetXRange() * 0.5f + GetMinXPos();
	float yOffs = GetYRange() * 0.5f + GetMinYPos();
	float zOffs = GetZRange() * 0.5f + GetMinZPos();

	float minXPos = max(minPoint0.m_XPos,minPoint1.m_XPos) + xOffs;
	float minYPos = max(minPoint0.m_YPos,minPoint1.m_YPos) + yOffs;
	float minZPos = max(minPoint0.m_ZPos,minPoint1.m_ZPos) + zOffs;

	float maxXPos = min(maxPoint0.m_XPos,maxPoint1.m_XPos) + xOffs;
	float maxYPos = min(maxPoint0.m_YPos,maxPoint1.m_YPos) + yOffs;
	float maxZPos = min(maxPoint0.m_ZPos,maxPoint1.m_ZPos) + zOffs;
			
	// Focus as new field boundaries=> set seedpoint to center and
	// recalculate field drawing "window".
	if(m_RecalcType == DEF_NEWBOUND_INDEX)
	{
		m_CurrentField->SetWinMinMaxPos(CPoint3D(minXPos,minYPos,minZPos),
										CPoint3D(maxXPos,maxYPos,maxZPos));
		m_SeedPCross.SetPosition(CPoint3D());
	}
	// Set seedpoint cross to middle of seedpoint region.
	else if(m_RecalcType == DEF_SEEDREG_INDEX)
		m_SeedPCross.SetPosition(CPoint3D(minXPos + (maxXPos - minXPos ) * 0.5f - xOffs,
										  minYPos + (maxYPos - minYPos ) * 0.5f - yOffs,
										  minZPos + (maxZPos - minZPos ) * 0.5f - zOffs));
	
	// Update document minmax-position.
	UpdateMinMaxPos();

	// Set seedpoint region to new boundaries.
	m_CurrentField->SetSeedMinMaxPos(CPoint3D(minXPos,minYPos,minZPos),
									 CPoint3D(maxXPos,maxYPos,maxZPos));

}

void CFlow3DDoc::OnOpenData() 
{
	if(m_CurrentField != NULL)
	{
		// TODO: Add your command handler code here
		CFileDialog fileDialog(TRUE);
	
		CVector3D vec;

		if(fileDialog.DoModal() == IDOK)
		{
			CString  filename = fileDialog.GetPathName();
			CFile file(filename, CFile::modeRead);

			// Load vectors and additional information.
			MyProgressDialog progressDialog;
			
			POSITION viewpos = GetFirstViewPosition();
			CView* pFirstView = GetNextView( viewpos );
			progressDialog.Create(IDD_PROGRESSDIALOG,pFirstView);
			progressDialog.InitProgressDialog(DEF_LOADDATA_STR,
											  DEF_SAVELOADDATA_MAX);
			progressDialog.ShowWindow(SW_SHOW);

			float scalar = 0.0f;
			char byt[sizeof(float)];
			float v[3];
			int u = 0;

			for(int z = 0; z < m_CurrentField->GetZDim(); z++)
			{
				progressDialog.m_Progress.StepIt();
				for(int y = 0; y < m_CurrentField->GetYDim(); y++)
					for(int x = 0; x < m_CurrentField->GetXDim(); x++)
					{
						CSamplePoint *sp = m_CurrentField->GetVField(x,y,z);
			
						for(int k=0;k<3;k++)
						{
							for(int j=0;j<sizeof(float);j++)
								file.Read(&byt[sizeof(float)-1-j],sizeof(char));

							v[k] = *((float *)byt);
						}

						vec.m_XComp = v[0];
						vec.m_YComp = v[1];
						vec.m_ZComp = v[2];

						sp->SetVector(vec);
						
						// Get block size.
						if((u == 0) && (sp->GetMagnitude() == 0.0f))
						{
							m_BlMinPos = *sp->GetPos();
							u ++;
						}
						else if((u == 1) && (sp->GetMagnitude() == 0.0f))
						{
							m_BlMaxPos = *sp->GetPos();				
						}

						for(int i = 0; i < CSamplePoint::GetAdditionalSize(); i++)
						{
							 for(int j=0;j<sizeof(float);j++)
								file.Read(&byt[sizeof(float)-1-j],sizeof(char));

							scalar = *((float *)byt);
							sp->SetAdditional(i,scalar);
						}
					}
			}	

			m_CurrentField->CalcMinMaxScalars();
			progressDialog.DestroyWindow();

		}
	}
	else
		MessageBox(NULL,"No vectorfield geometry specified","Error",MB_OK);
}

void CFlow3DDoc::OnSaveDataAs() 
{
	// TODO: Add your command handler code here
	CFileDialog fileDialog(FALSE);

	if(fileDialog.DoModal() == IDOK)
	{
		CString  filename = fileDialog.GetPathName();
		CFile file(filename, CFile::modeCreate | CFile::modeWrite );

		CVector3D vec;

		if(m_CurrentField != NULL)
		{
			MyProgressDialog progressDialog;
			POSITION viewpos = GetFirstViewPosition();
			CView* pFirstView = GetNextView( viewpos );
			progressDialog.Create(IDD_PROGRESSDIALOG,pFirstView);
			progressDialog.InitProgressDialog(DEF_SAVEDATA_STR,
											  DEF_SAVELOADDATA_MAX);
			progressDialog.ShowWindow(SW_SHOW);

			char byt[sizeof(float)];
			char *revByt;
			
			// Storing vector and additional information.
			for(int z = 0; z < m_CurrentField->GetZDim(); z++)
			{
				progressDialog.m_Progress.StepIt();

				for(int y = 0; y < m_CurrentField->GetYDim(); y++)
					for(int x = 0; x < m_CurrentField->GetXDim(); x++)
					{
						// Storing vector information.
						m_CurrentField->GetVField(x,y,z)->GetVector(vec);
			
						for(int k=0;k<3;k++)
						{
							switch(k)
							{
								case 0:
								revByt = (char *)(&vec.m_XComp);
								break;
								case 1:
								revByt = (char *)(&vec.m_YComp);
								break;
								case 2:
								revByt = (char *)(&vec.m_ZComp);
								break;
							}
							
							// Reverse byte-order.
							for(int j=0;j<sizeof(float);j++)
								byt[sizeof(float)-1-j] = revByt[j];
				
							file.Write(byt, sizeof(float));
						}

						// Now storing addit. scalar information.
						for(int i=0;i<CSamplePoint::GetAdditionalSize();i++)
						{
							float scalar = 
								m_CurrentField->GetVField(x,y,z)->
									GetAdditional(i);
							
							revByt = (char *)(&scalar);

							// Reverse byte-order.
							for(int j=0;j<sizeof(float);j++)
								byt[sizeof(float)-1-j] = revByt[j];
				
							file.Write(byt, sizeof(float));
						}
					}	
			}
			progressDialog.DestroyWindow();
		}
	}
}

void CFlow3DDoc::InitRegions()
{
	// Init scalar regions: 
	// region with index DEF_REGIONS_NUM = default properties.
	GLubyte ambientRGB[3];
	GLubyte diffuseRGB[3];
	GLubyte specularRGB[3];

	ambientRGB[0] = DEF_MAT_AMB_R;
	ambientRGB[1] = DEF_MAT_AMB_G;
	ambientRGB[2] = DEF_MAT_AMB_B;

	diffuseRGB[0] = DEF_MAT_DIF_R;
	diffuseRGB[1] = DEF_MAT_DIF_G;
	diffuseRGB[2] = DEF_MAT_DIF_B;

	specularRGB[0] = DEF_MAT_SPEC_R;
	specularRGB[1] = DEF_MAT_SPEC_G;
	specularRGB[2] = DEF_MAT_SPEC_B;

	for(int i=0;i<DEF_REGPROP_NUM;i++)
	{
		m_RegProp[i].material = 
			CStreamlineMaterial(ambientRGB,diffuseRGB,specularRGB,DEF_OPACITY);

		m_RegProp[i].res = DEF_RES_MAX;
		m_RegProp[i].pos = DEF_REGIONPOS_MIN;
		m_RegProp[i].lineWidth = DEF_LINEWIDTH;
		UpdateResIndex(m_RegProp[i]);
	}

	// Set position values of first 3 regions
	// So that per default 3 different regions are shown.
	m_RegProp[0].pos = DEF_REGION1_POS;
	m_RegProp[1].pos = DEF_REGION2_POS;
	m_RegProp[2].pos = DEF_REGION3_POS;

	GLubyte reg1Amb[4] = DEF_REGION1_AMB;
	GLubyte reg1Dif[4] = DEF_REGION1_DIF;
	GLubyte reg1Spec[4] = DEF_REGION1_SPEC;

	GLubyte reg2Amb[4] = DEF_REGION2_AMB;
	GLubyte reg2Dif[4] = DEF_REGION2_DIF;
	GLubyte reg2Spec[4] = DEF_REGION2_SPEC;

	GLubyte reg3Amb[4] = DEF_REGION3_AMB;
	GLubyte reg3Dif[4] = DEF_REGION3_DIF;
	GLubyte reg3Spec[4] = DEF_REGION3_SPEC;

	m_RegProp[0].material.SetAmbientRGB(reg1Amb);
	m_RegProp[0].material.SetDiffuseRGB(reg1Dif);
	m_RegProp[0].material.SetSpecularRGB(reg1Spec);
	m_RegProp[1].material.SetAmbientRGB(reg2Amb);
	m_RegProp[1].material.SetDiffuseRGB(reg2Dif);
	m_RegProp[1].material.SetSpecularRGB(reg2Spec);

	m_RegProp[2].material.SetAmbientRGB(reg3Amb);
	m_RegProp[2].material.SetDiffuseRGB(reg3Dif);
	m_RegProp[2].material.SetSpecularRGB(reg3Spec);
}

void CFlow3DDoc::InitContTransfProp()
{
	GLubyte minRgba[4] = DEF_MINRGBA;
	GLubyte interRgba[4] = DEF_INTERRGBA;
	GLubyte maxRgba[4] = DEF_MAXRGBA;

	// Set intermediate position in continuous transfer function
	// to default value.
	m_InterPos = DEF_INTERPOS;

	for(int i= 0; i < DEF_CONTTRANSFVALUES_NUM; i++)
	{
		m_ContTransfProp[i].res = DEF_RES_MAX;
		m_ContTransfProp[i].lineWidth = DEF_LINEWIDTH;
	}

	for(i = 0; i < 4; i++)
	{
		m_ContTransfProp[DEF_MINPOS_INDEX].rgba[i] = minRgba[i];
		m_ContTransfProp[DEF_INTERPOS_INDEX].rgba[i] = interRgba[i];
		m_ContTransfProp[DEF_MAXPOS_INDEX].rgba[i] = maxRgba[i];
	}

}
