// Flow3DDoc.h : interface of the CFlow3DDoc class
//
/////////////////////////////////////////////////////////////////////////////
#include "MyAlgoOptDialog.h"
#include "VectorField.h"
#include "StreamlineQueue.h"
#include "MyMaterialDialog.h"	// Added by ClassView
#include "Point3D.h"	// Added by ClassView
#include "Defines.h"
#include "StreamlinePiece.h"
#include "DSepGrid.h"
#include "MyAxisCross.h"
#include "MyHistoDialog.h"
#include "MyContTransfDialog.h"
#include "MyRegPropDialog.h"
#include "MyFocusContextDialog.h"

#if !defined(AFX_FLOW3DDOC_H__44A19073_6C34_494B_8782_6CB89A05907A__INCLUDED_)
#define AFX_FLOW3DDOC_H__44A19073_6C34_494B_8782_6CB89A05907A__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//!	Main document class of Flow3D application.
class CFlow3DDoc : public CDocument
{
protected: // create from serialization only
	CFlow3DDoc();
	DECLARE_DYNCREATE(CFlow3DDoc)

// Attributes
public:
	//! Properties of regions. Index DEF_REGIONS_NUM holds default region.
	RegProp m_RegProp[DEF_REGPROP_NUM];

	//! Properties of streamlines falling into context.
	ContextProp m_ContextProp;
	//! Continuous transfer function properties of streamlines.
	ContTransfProp m_ContTransfProp[DEF_CONTTRANSFVALUES_NUM];

	//! End tapering line width table.
	float * m_EndTapTable;

	//! Magic volume = boundaries of focus.
	MyObject *m_MagicVolume;

	//! Minimal scalar values
	float *m_MinScalars;
	//! Maximal scalar values.
	float *m_MaxScalars;

	//! Maximal position value of samplepoints.
	CPoint3D m_MaxPos;
	//! Minimal position value of samplepoints.
	CPoint3D m_MinPos;

	//! Current length of transparency functions(only haze linear, sinus).
	int m_PeriodLength;

	//! Streamline start position in animation table.
	POSITION *m_AnimStartTbl;
	//! Streamline end position in animation table.
	POSITION *m_AnimEndTbl;

	//! Indicates if streamlines should be animated.
	bool m_IsAnimation;
	//! Indicates if arrows should be drawn.
	bool m_IsArrows;

	//! Width of alpha table.
	int m_AlphaTblWidth;

	//! Table which holds opacity values for transparency function.
	GLubyte *m_AlphaTable;

	//! Magic volume type index.
	int m_MagicVolType;

	//! Boundaries of field.
	MyRectangle m_Boundaries;

	//! Minimal position of block(block-dataset).
	CPoint3D m_BlMinPos;

	//! Maximal position of block(block-dataset).
	CPoint3D m_BlMaxPos;

	//! Z-offset value of block(block-dataset).
	float m_BlZOffs;

	//! Block material(block-dataset).
	CStreamlineMaterial m_BlMaterial;

	//! Is seedpoint-placing mode?
	bool m_IsPlaceSeedPoint;

	//! If in focus/context modus, this member defines the recalculation type.
	int m_RecalcType;

	//! Animation speed of streamlines.
	int m_AnimSpeed;

	//! Animation type.
	int m_AnimType;

// Operations
public:
	//! Calculates new field boundaries if field is recalculated.
	void CalcFieldMinMaxPos();
	//! Updates animation index table.
	void UpdateAnimIndex();
	//! Table of streamline list positions.
	void CreateAnimIndexTbl();
	//! Tests streamlines if they are inside focus.
	void DoInsideTest();
	//! Updates context streamline properties.
	void UpdateContextProp();
	//! Returns type of additional scalar info representation.
	int GetAddInfoRepr();
	//! Sets type of additional scalar info representation.
	void SetAddInfoRepr(int repr);
	//! Creates array with various minimal and maximal scalar values.
	void CreateMinMaxScalars();

	//! Creates table for line end tapering values.
	float *CreateEndTapTable(int width);

	//! Returns pointer to vectorfield.
	CVectorField *GetCurrentField();

	//! Returns number of candidates per samplepoints.
	int GetCaPerSP();
	//! Returns maximal number of calculated streamlines.
	int GetResNum();

	//! Returns maximal number of iterations.
	int GetMaxIterations();

	//! Returns dt.
	float GetDt();
	//! Returns dTest.
	float GetDTest();
	//! Returns dSep(in % of vectorfield x-range).
	float GetDSepPerc();

	//! Returns maximal x-position of current vectorfield.
	float GetMaxXPos();
	//! Returns maximal y-position of current vectorfield.
	float GetMaxYPos();
	//! Returns maximal z-position of current vectorfield.
	float GetMaxZPos();
	//! Returns minimal x-position of current vectorfield.
	float GetMinXPos();
	//! Returns minimal y-position of current vectorfield.
	float GetMinYPos();
	//! Returns minimal z-position of current vectorfield.
	float GetMinZPos();

	//! Returns range of vectorfield in x-direction.
	float GetXRange();
	//! Returns range of vectorfield in y-direction.
	float GetYRange();
	//! Returns range of vectorfield in z-direction.
	float GetZRange();

	//! Returns minimal distance between samplepoints in percent of field x-range
	float GetMinDistPerc();

	//! Returns number of streamlines.
	int GetStreamlinesNum();

	//! Returns number of displayed samplepoints.
	long GetDisplSamplePNum();

	//! Returns number of samplepoints.
	long GetSamplePointsNum();

	//! Returns calculation time of streamlines in seconds.
	long GetCalculationTime();

	//! Returns flag which indicates if correct rendering order should be used.
	bool GetIsCorrectOrder();

	//! Returns flag which indicates if tapering should be used.
	bool GetIsTapering();

	//! Returns flag which indicates if display lists should be used.
	bool GetUseDisplayLists();

	
	//! Returns flag which indicates if additional information should be displayed.
	bool GetIsAddInfo();

	//! Returns flag which indicates if context streamlines should be cut in focus.
	bool GetIsCutLines();

	//! Returns flag which indicates if magic volume should be shown.
	bool GetIsMagicVol();

	//! Returns flag which indicates if magic volume should be shown.
	bool GetIsIncremential();
	
	//! Returns flag which indicates if magic volume should be shown.
	bool GetIsCalcIncr();

	//! Returns flag which indicates if focus/context modus should be shown.
	bool GetIsFocusContext();

	//! Returns intermediate position from continuous transfer function.
	float GetInterPos();

	//! Returns additional info type
	int GetAddInfoType();
	
	//! Returns scalar value with specified index;
	float GetScalar(int index, CStreamlinePiece *piece);
	
	//! Sets flag which indicates if tapering should be used.
	void SetIsTapering(bool isTapering);

	//! Sets flag which indicates if context streamlines should be cut in focus.
	void SetIsCutLines(bool isCutLines);

	//! Sets flag which indicates if correct rendering order should be used.
	void SetIsCorrectOrder(bool isCorrectOrder);

	//! Sets flag which indicates if additional information should be displayed.
	void SetIsAddInfo(bool isAddInfo);

	//! Sets minimal distance between samplepoints in percent of field x-range
	void SetMinDistPerc(float minDistPerc);

	//! Sets flag which indicates if display lists should be used.
	void SetUseDisplayLists(bool useDisplayLists);

	
	//! Sets additional info type
	void SetAddInfoType(int addInfoType);

	//! Sets middle pos from color code transfer function.
	void SetInterPos(float interPos);

	//! Updates min/max position values of vectorfield boundaries.
	void UpdateMinMaxPos();

	//! Sets flag which indicates if magic volume should be shown.
	void SetIsMagicVol(bool isMagicVol);

	//! Sets flag which indicates if focus/context modus should be shown.
	void SetIsFocusContext(bool isFocusContext);

	//! Computes thickness coefficient of streamline for tapering effect.
	float ThicknessCoeff(CSamplePoint *samplePoint);

	//! returns list of streamlines.
	CObList * GetStreamlines();
	
	//! Calculates min + max scalars of streamline pieces for add. info.
	void CalcMinMaxScalars();

	//! Returns pointer to pieces list ordered in x-direction upwards.
	CObList * GetXList();
	//! Returns pointer to pieces list ordered in y-direction upwards.
	CObList * GetYList();
	//! Returns pointer pieces list ordered in z-direction upwards.
	CObList * GetZList();
	
	//! Returns axis cross for specifying first seedpoint of algorithm
	MyAxisCross * GetSeedPCross();
	
	//! Inserts new piece into x-direction-list: list sorted if sort = true.
	POSITION InsertXList(CStreamlinePiece *piece, POSITION pos, bool sort);
	//! Inserts new piece into x-direction-list: list sorted if sort = true.
	POSITION InsertYList(CStreamlinePiece *piece, POSITION pos, bool sort);
	//! Inserts new piece into x-direction-list: list sorted if sort = true.
	POSITION InsertZList(CStreamlinePiece *piece, POSITION pos, bool sort);

	//! Updates pieces lists.
	void UpdatePiecesLists();

	//! Updates one streamline piece with current properties and given scalar values.
	void UpdatePiece(CStreamlinePiece *piece, float *scalarVals);

	//! Calculates new streamlines.
	void Recalc();

	//! Creates new vector field with analytic equations.
	void CreateField();

	//! Inserts pieces of new streamlines into x,y and z-list and inits values
	void InsertPiecesLists(CStreamline *streamline);

	//! Resorts streamline pieces lists and inits values.
	void ResortPiecesLists();

	//! Deletes pieces lists.
	void DeletePiecesLists();

	//! Deletes all streamlines.
	void DeleteStreamlines();

	//! Returns type of interpolator.
	int GetIntegratorType();

	//! Returns type of integrator.
	int GetInterpolatorType();

	//! Updates resolution index on specifie region index.
	void UpdateResIndex(RegProp &prop);

	//! Returns width(=size) of end tapering table.
	int	GetEndTapWidth();
	//! Sets width(=size) of end tapering table.
	void SetEndTapWidth(int width);

	//! Indicates if end tapering should be used.
	bool GetIsEndTap();
	//! Sets flag if end tapering should be used.
	void SetIsEndTap(bool isEndTap);

	//! Calculates end tapering indeces for streamline pieces.
	int CalcEndTapIndex(int index, int width, int lineLength);
	
	//! Calculates end tapering and arrow position for every streamline pieces.
	void CalcArrow_EndTap();

	//! Returns flag if streamline pieces should be drawn to screen.
	bool IsDraw(bool &isInside, CStreamlinePiece *piece, int res, 
				int focusRes, int contextRes);

	//! Returns minimal and maximal scalar value for scalar of index i.
	void GetMinMaxScalar(int index, float &minScalar, float &maxScalar);

	// Sets property values for continuous transfer function.
	void SetContTransfProp(int index, ContTransfProp &prop);
	// Sets property values for continuous transfer function.
	void GetContTransfProp(int index, ContTransfProp &prop);

	//! Calculates which pieces of the streamline should be drawn.
	void CalcIsDraw();

	// Initialises scalar regions.
	void InitRegions();

	// Initialises continuous scalar transfer function properties.
	void InitContTransfProp();

	virtual ~CFlow3DDoc();

	
private:
	int m_AnimTblWidth;
	int m_MaxResIndex;

	int m_AddInfoRepr;
	bool m_IsEndTap;
	int m_EndTapWidth;
	CObList m_LinePieces;
	CObList m_Streamlines;

	int m_ResNum;
	int m_CalcResNum;

	MyAxisCross m_SeedPCross;
	
	bool m_IsCalcIncr;
	bool m_IsIncremential;

	bool m_IsCutLines;

	int m_MaxIterations;
	int m_IntegratorType;
	int m_InterpolatorType;
	int m_CaPerSP;

	float m_DSepPerc;
	float m_DTest;
	float m_Dt;
	
	CDSepGrid m_DSepGrid;

	//! Is in focus/context modus?
	bool m_IsFocusContext;

	//! Should magic volume be shown?
	bool m_IsMagicVol;

	//! Minimal distance between samplepoints in % of field x-range.
	float m_MinDistPerc;
	
	//! Index value for type of additional info.
	int m_AddInfoType;

	//! Intermediate position from continuous transfer function.
	float m_InterPos;
	
	//! Should texture regions be used for additional info?
	bool m_IsRegions;
	//! Should additional info be displayed?
	bool m_IsAddInfo;
	//! Should continuous transfer function be used for additional info?
	bool m_IsContTransf;

	//! Number of samplepoints displayed on screen.
	long m_DisplSamplePNum;
	//! Time of streamline calculation.
	long m_CalculationTime;
	//! Number of samplepoints.
	long m_SamplePointsNum;
	//! Number of streamlines.
	int m_StreamlinesNum;

	//! Should display lists be used?
	bool m_UseDisplayLists;
	//! Should correct rendering order be used?
	bool m_IsCorrectOrder;

	//! Should tapering be used?
	bool m_IsTapering;

	//! Parameter 1 influencing field sampling.
	float m_Param1;
	//! Parameter 1 influencing field sampling.
	float m_Param2;
	//! Parameter 1 influencing field sampling.
	float m_Param3;
	//! Z-dimension of vectorfield.
	int m_ZDim;
	//! Y-dimension of vectorfield.
	int m_YDim;
	//! X-dimesion of vectorfield.
	int m_XDim;
	//! Index value used to specify type of field sampling equations.
	int m_SamplerType;
	//! Should additional info like pressure be sampled?
	bool m_IsSampleAddInfo;

	// Current vectorfield.
	CVectorField *m_CurrentField;
	
	//! Streamline pieces list ordered in x-direction upwards.
	CObList m_ZList;
	//! Streamline pieces list ordered in y-direction upwards.
	CObList m_YList;
	//! Streamline pieces list ordered in z-direction upwards.
	CObList m_XList;

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CFlow3DDoc)
	public:
	virtual BOOL OnNewDocument();
	virtual void Serialize(CArchive& ar);
	virtual void DeleteContents();
	//}}AFX_VIRTUAL

// Implementation
public:
	
		
#ifdef _DEBUG
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CFlow3DDoc)
	afx_msg void OnOptionsAlgorithm();
	afx_msg void OnSetFieldSamplingOptions() ;
	afx_msg void OnSetResolutionOptions();
	afx_msg void OnOpenData();
	afx_msg void OnSaveDataAs();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()

};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_FLOW3DDOC_H__44A19073_6C34_494B_8782_6CB89A05907A__INCLUDED_)
