// Flow3DView.h : interface of the CFlow3DView class
//
/////////////////////////////////////////////////////////////////////////////
#include "HistoData.h"	// Added by ClassView
#include "Streamline.h"
#include "Vector3D.h"	// Added by ClassView
#include "Point3D.h"	// Added by ClassView
#include "Defines.h"
#include "MyCube.h"
#include "TextureGenerator.h"
#include "MyProgressDialog.h"	// Added by ClassView
#include "MyRectangle.h"
#include "MyAxisCross.h"
#include "MySphere.h"
#include "MyHistoDialog.h"

#include "gl/gl.h"
#include "gl/glu.h"
#include "gl/glut.h"

#if !defined(AFX_FLOW3DVIEW_H__8D49CA2C_4BB9_4D56_A76B_282A9CF7747B__INCLUDED_)
#define AFX_FLOW3DVIEW_H__8D49CA2C_4BB9_4D56_A76B_282A9CF7747B__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//! Main view class for Flow3D application.
class CFlow3DView : public CView
{
protected: // create from serialization only
	CFlow3DView();
	DECLARE_DYNCREATE(CFlow3DView)

// Attributes
public:
	CFlow3DDoc* GetDocument();
	
// Operations
public:
		
	virtual ~CFlow3DView();
// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CFlow3DView)
	public:
	virtual void OnDraw(CDC* pDC);  // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	protected:
	virtual void OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint);
	//}}AFX_VIRTUAL

// Implementation
				
private:
	//! Initialise focus/context graphical objects.
	void InitFoCoObj();
	//! Initialises block parameters(only block dataset).
	void InitBlockParams();

	//! Initialises variables used for mouse-move routine.
	void InitOldMembers();

	//! Replaces block texture with another one.
	void ReplaceBlTex();

	//! Creates opengl-block texture.
	void CreateBlTex();

	//! Draws block from block-dataset.
	void DrawBlock();

	//! Inits gl-state.
	void GLInit();
	//! Inits histogram dialog.
	void InitHistoDialog(int scalarType);

	//! Returns streamline pieces-list representing current view direction.
	CObList *GetPiecesList(bool &isReverse, int direction);
	//! Percentage of current field size relative to full field size.
	float CalcSizePerc();
	//! Inits graphical objects like boundaries or magic volume.
	void InitObjects();
	
	//! Centers focus to be center of window.
	void CenterFocus();

	//! Draws cone-shaped arrows on end of streamlines.
	void DrawCone(CSamplePoint *sp, CStreamlineMaterial &mat, 
				  GLubyte *rgba, float base, float height);

	//! Draws vector field with arrows.
	void DrawVField();
	//! Sets depth cueing parameters.
	void SetDepthCueing();
	//! Draws bar indicating current animation speed.
	void DrawAnimSpeedBar(int xPos, int yPos, int width, GLubyte *color);
		
	//! Draws legend with important information.
	void DrawLegend();

	//! Calculates new size of magic volume according to mouse movement vector vec.
	void CalcNewSize(float &size,float &radius,float &length,
					 float &width, float &height, CVector3D vec);

	//! Calculates new position according to mouse movement vector vec.
	CPoint3D CalcNewPos(CPoint3D oldPos, CVector3D vec);

	//! Rotates field according to rotangle and rotaxis.
	void RotateField();

	//! Draws xyz-axis lines for lengend.
	void DrawLines(GLint xPos, GLint yPos, GLint length, 
				   GLint width, GLubyte opacity);

	//! Draws text from string s(newlines are also possibl).
	void DrawText(GLint xPos, GLint yPos, char* s, GLubyte *rgba);

	//! Draws legend with colors of main coordinate axis.
	void DrawAxisLegend(GLint xPos, GLint yPos, GLubyte *rgba);

	//! Creates table of alpha values for the animation.
	GLubyte *CreateAlphaTable();

	//! Returns frames per seconds.
	int CalcFPS();

	//! Initialises member variables and opengl-parameters
	void Init();

	//! Sets up pixel format in gl-initialisation.
	BOOL bSetupPixelFormat(void);
	//! Sets up texture matrix with current view vector.
	void SetTextureMatrix();

	//! modulo transparency function for streamlines
	GLubyte ModFunc(int i, int n);

	//! sinus transparency function for streamlines
	GLubyte SinFunc(int i, int n);

	//! Computes main direction. 
	int GetMainDirection();

	//! Main drawing procedure: Calls DrawCube, DrawStreamlines, DrawAxis etc.
	void DrawScene();
	
	//! Draw computed streamlines.
	void DrawStreamlines(int direction);

	//! Replaces texture maps with current texture maps.
	void ReplaceTextures();

	//! Creates texture maps for the first time.
	void CreateTextures();

	//! Sets matrix to current rotation matrix
	void SetRotMatrix(float *matrix);

	//! Clamps value if < min or => max.
	static void Clamp(float &value, float min, float max, float newVal1, float newVal2);

	//! Indicates if two floats have changed.
	static bool HasChanged(float oldVal, float val);

	//! Indicates if two rgba values have changed.
	static bool HasChanged(GLubyte *oldRgba, GLubyte *rgba);

	//! Is a key pressed?
	bool IsKeyPressed();

	//! Is field rotated with left button?
	bool IsLButtonRot();

	//! Depth coding intensity.
	int m_DeIntensity;

	//! Pointer to document.
	CFlow3DDoc *doc;

	//! Histogram dialog.
	MyHistoDialog m_HistoDialog;
	
	//! Rotation x-angle of "view direction" dialog.
	float m_ZAngle;
	//! Rotation y-angle of "view direction" dialog.
	float m_YAngle;
	//! Rotation z-angle of "view direction" dialog.
	float m_XAngle;

	//! Vectorfield-var for display list.
	int m_TheVField;

	//! Old minimum field xyz-position.
	CPoint3D m_OldMinPos;
	
	//! Old maximum xyz-position.
	CPoint3D m_OldMaxPos;

	//! Old animation speed.
	int m_OldAnimSpeed;
	//! Old period length.
	int m_OldPeriodLength;
	//! Old depth cuing intensity.
	int m_OldDeIntensity;
	//! Distance from z-axis.
	float m_ZDist;
	//! Old distance from z-axis.
	float m_OldZDist;
	//! Absorption factor of particle haze.
	GLubyte m_OldAbsorption;
	//! Old context properties.
	ContextProp m_OldContextProp;
	//! Old block-zoffs(block dataset only)
	float m_OldBlZOffs;

	//! Background rgba-color.
	GLfloat m_BKColor[4];
	//! Depth rgb-color.
	GLubyte m_DepthColor[3];

	//! Position where right mouse button was pressed last time.
	CPoint m_RButtonDownPos;
	//! Position where left mouse button was pressed last time.
	CPoint m_LButtonDownPos;
	//! Current mouse position.
	CPoint m_CurrentMousePos;
	//! Position where left mouse button was double clicked last time.
	CPoint m_RButDblClkPos;

	//! Old seedpoint-placing cross.
	MyAxisCross m_OldSeedPCross;

	//! Old magic volume.
	MyObject *m_OldMagicVolume;

	//! Main axis.
	MyAxisCross m_Axis;
	//! Axis for placing magic volume in focus-context mode.
	MyAxisCross m_FoContextCross;

	//! Absorption from one streamline point to the next.
	GLubyte m_Absorption;
	
	//! Current rotation axis.
	CVector3D m_RotAxis;
	//! Current rotation angle.
	float m_RotAngle;

	//! Current size of gaps between crossing streamlines.
	float m_GapSize;

	//! Depth coding type.
	int m_DepthType;

	//! Arrow period length.
	int m_ArPeriod;

	//! Arrow drawing type.
	int m_ArType;

	//! Streamline arrow width.
	float m_ArrowWidth;
	//! Streamline arrow length.
	float m_ArrowLength;

	// State variables.
	bool m_IsBusy;
	bool m_IsShowLegend;
	bool m_IsKeyPressed[DEF_KEYS_NUM];
	
	bool m_RButtonDown;
	bool m_IsRButtonDown;
	bool m_IsLButtonDown;

	bool m_IsDepthCueing;
	bool m_IsDrawBoundaries;
	bool m_IsRotating;
	bool m_IsAntiAliasing;

	bool m_IsGap;
	bool m_HasFocus;
	bool m_Modified;
	bool m_IsTranspFunc;
	bool m_IsShowAxis;
	bool m_IsShowBlock;
	bool m_IsTex;

	//! Index value for light type(headlight,pointlight).
	int m_LightType;
	//! Index value for type of transparency function(sin, haze lin, haze exp).
	int m_TranspFuncType;

	//! Current diffuse color component correction coefficient.
	float m_DiffuseCorr;
	//! Current glossiness factor of streamline material.
	float m_Glossiness;

	//! Index of display lists representing the streamlines.
	GLuint m_TheStreamlines;
	
	//! Index of textures. Index DEF_REGIONS_NUM holds default texture map.
	GLuint texName[DEF_REGPROP_NUM+1];

	//! Index of block-texture(block-dataset only).
	GLuint blTexName;

	//! Light direction vector.
	CVector3D m_Light;

	//! View direction vector.
	CVector3D m_View;

	//! Texture generator generates texture maps for illuminated streamlines.
	CTextureGenerator m_TexGenerator;

	//! Current context.
	CClientDC * m_pDC;

	//! Window rectangle.
	CRect m_oldRect;

	//! Sinus table.
	static CSinTable m_SinTable;

#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CFlow3DView)
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnSize(UINT nType, int cx, int cy);
	afx_msg void OnDestroy();
	afx_msg BOOL OnEraseBkgnd(CDC* pDC);
	afx_msg void OnShowHideBoundaries();
	afx_msg void OnEnableDisableDepthCueing();
	afx_msg void OnTimer(UINT nIDEvent);
	afx_msg void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnRButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnLButtonDblClk(UINT nFlags, CPoint point);
	afx_msg void OnPlaceSeedPoint();
	afx_msg void OnSetAddInfo();
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnSetMaterial();
	afx_msg void OnSetHighlight();
	afx_msg void OnSetDiffuseCorrection();
	afx_msg void OnShowHideAddInfo();
	afx_msg void OnSetLightType();
	afx_msg void OnUpdateEnableDisableDepthCueing(CCmdUI* pCmdUI);
	afx_msg void OnUpdatePlaceSeedPoint(CCmdUI* pCmdUI);
	afx_msg void OnUpdateShowHideAddInfo(CCmdUI* pCmdUI);
	afx_msg void OnUpdateShowHideBoundaries(CCmdUI* pCmdUI);
	afx_msg void OnSetLineStyle();
	afx_msg void OnShowHideAxis();
	afx_msg void OnUpdateShowHideAxis(CCmdUI* pCmdUI);
	afx_msg void OnStatistics();
	afx_msg void OnEnableDisableAnimation();
	afx_msg void OnUpdateEnableDisableAnimation(CCmdUI* pCmdUI);
	afx_msg void OnSetFocus(CWnd* pOldWnd);
	afx_msg void OnKillFocus(CWnd* pNewWnd);
	afx_msg void OnSetPerformanceOptions();
	afx_msg void OnShowHideMagicVolume();
	afx_msg void OnUpdateShowHideMagicVolume(CCmdUI* pCmdUI);
	afx_msg void OnSetFocusContextOptions();
	afx_msg void OnStartEndFocusContextMode();
	afx_msg void OnUpdateStartEndFocusContextMode(CCmdUI* pCmdUI);
	afx_msg void OnRButtonDblClk(UINT nFlags, CPoint point);
	afx_msg void OnRButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnShowHideLegend();
	afx_msg void OnUpdateShowHideLegend(CCmdUI* pCmdUI);
	afx_msg void OnSetBackground();
	afx_msg void OnZoomIn();
	afx_msg void OnResetSize();
	afx_msg void OnShowHistoDifOrient();
	afx_msg void OnShowHistoOrientX();
	afx_msg void OnShowHistoOrientY();
	afx_msg void OnShowHistoOrientZ();
	afx_msg void OnShowHistoPosX();
	afx_msg void OnShowHistoPosY();
	afx_msg void OnShowHistoPosZ();
	afx_msg void OnShowHistoVelocity();
	afx_msg void OnShowHistoDifVelocity();
	afx_msg void OnSetDepthCoding();
	afx_msg void OnSetAnimationOptions();
	afx_msg void OnSetViewDirection();
	afx_msg void OnSetBlockParameters();
	afx_msg void OnShowHideBlock();
	afx_msg void OnUpdateShowHideBlock(CCmdUI* pCmdUI);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

#ifndef _DEBUG  // debug version in Flow3DView.cpp
inline CFlow3DDoc* CFlow3DView::GetDocument()
   { return (CFlow3DDoc*)m_pDocument; }
#endif

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_FLOW3DVIEW_H__8D49CA2C_4BB9_4D56_A76B_282A9CF7747B__INCLUDED_)
