// SamplePoint.h: interface for the CSamplePoint class.
//
//////////////////////////////////////////////////////////////////////
#include "Vector3D.h"
#include "Point3D.h"

#if !defined(AFX_SAMPLEPOINT_H__60A73276_047C_4915_8F63_653EEA1F5212__INCLUDED_)
#define AFX_SAMPLEPOINT_H__60A73276_047C_4915_8F63_653EEA1F5212__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//! A class that represents a samplepoint in a vectorfield. 
/*!	CSamplePoint consists of a position, a vector and an array of additional
	information like pressure.
	The vector is stored as unit vector + magnitude, therefore directly operating
	with the unit vector is fastest and requires no additional computation.
	Also the quadratic distance to other samplepoints can be stored.
*/

class CSamplePoint: public CObject  
{
public:
	float * GetAdditional();
	void SetAdditional(float *additional);
	//! Returns unit vector x-component.
	float GetUnitZVec();
	//! Returns unit vector y-component.
	float GetUnitYVec();
	//! Returns unit vector z-component.
	float GetUnitXVec();
	//! Returns unit vector.
	CVector3D *GetUnitVec();
	//! Sets unit vector.
	void SetUnitVec(CVector3D &unitVec);
	//! Sets vector length.
	void SetMagnitude(float magnitude);
	//! Returns length of vector.
	float GetMagnitude();

	//! Overloaded operator '='.
	CSamplePoint& operator=(const CSamplePoint &samplePoint);

	//! Copy-constructor.
	CSamplePoint(const CSamplePoint &samplePoint);

	//! Sets values of this samplepoint to the values of parameter.
	static int GetAdditionalSize();
	//! Sets size of additional information-array
	static void SetAdditionalSize(int size);

	//! Returns x-component of vector.
	float GetZVec();
	//! Returns y-component of vector.
	float GetYVec();
	//! Returns z-component of vector.
	float GetXVec();
	
	CSamplePoint();
	//! Constructor with position, vector and additional-array as parameters
	/*!
		For the internal representation the magnitude of the vector is computed
		and the vector is normalized.
	*/
	CSamplePoint(CPoint3D &pos3D, CVector3D &vector3D,float *additional); 
	//! Constructor with position, unit vector and additional-array as parameters
	/*!
		Because of the internal representation this representation requires no
		additional computation and is the faster constructor
	*/
	CSamplePoint(CPoint3D &pos3D, CVector3D &unitVec3D,float magnitude,
				 float *additional); 	    

	virtual ~CSamplePoint();

	//! Sets position of samplepoint.
	void SetPos(CPoint3D &pos3D);

	//! Sets vector. 
	/*!
		For the internal representation the magnitude of the vector is computed
		and the vector is normalized.
	*/
	void SetVector(CVector3D &vector3D);

	//! Sets additional information (pressure,velocity,...).
	void SetAdditional(int index, float value);

	//! Gets vector. Requires multiplication unit vector * vector magnitude.
	void GetVector(CVector3D &vec);

	//! Returns x-position of the samplepoint.
	float GetXPos();
	//! Returns y-position of the samplepoint.
	float GetYPos();
	//! Returns z-position of the samplepoint.
	float GetZPos();

	//! Returns element i in additional information array 
	float GetAdditional(int i);
	
	//! Returns quadratic minimal distance to other samplepoints
	float GetQuadMinDist();

	//! returns position of samplepoint as CPoint3D object
	CPoint3D *GetPos();

	//! Returns euclidian distance to another samplepoint
	float Distance(CSamplePoint *samplePoint);
	
	//! Returns quadratic euclidian distance to another samplepoint
	float GetQuadDistance(CSamplePoint *samplePoint);

	//! Computes minimal quadratic euclidian distance to an array of samplepoints, false if < minDist
	bool CSamplePoint::ComputeQuadMinDist(CObList *samplePoints,float minDist);
	
private:
	
	//! Position of samplepoint.
	CPoint3D m_Pos;
	//! Vector.
	CVector3D m_Vector;
	//! Stores additional scalar information.
	float *m_Additional;

	//! Quadratic minimal distance.
	float m_QuadMinDist;

	//! Length of this vector.
	float m_Magnitude;

	//! Stores number of additional information.
	static int m_AdditionalSize;

};



#endif // !defined(AFX_SAMPLEPOINT_H__60A73276_047C_4915_8F63_653EEA1F5212__INCLUDED_)

