// Vector3D.cpp: implementation of the CVector3D class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Flow3D.h"
#include "Vector3D.h"
#include "math.h"
#include "Defines.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

CSinTable CVector3D::m_SinTable(DEF_ANGLE_0F, DEF_ANGLE_360F, DEF_SINTABLE_WIDTH);
CCosTable CVector3D::m_CosTable(DEF_ANGLE_0F, DEF_ANGLE_360F, DEF_COSTABLE_WIDTH);
CVector3D CVector3D::m_HVec[3];

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CVector3D::CVector3D()
{
	m_XComp = 0.0f; 
	m_YComp = 0.0f; 
	m_ZComp = 0.0f;
}

CVector3D::CVector3D(float xComp, float yComp, float zComp)
{ 
	m_XComp = xComp; 
	m_YComp = yComp; 
	m_ZComp = zComp;
};

CVector3D::~CVector3D()
{

}

float CVector3D::DotProduct(CVector3D &vector3D)
{
	return m_XComp*vector3D.m_XComp + 
		   m_YComp*vector3D.m_YComp +
		   m_ZComp*vector3D.m_ZComp;
}


void CVector3D::CrossProduct(CVector3D &source, CVector3D &dest)
{
	dest.m_XComp = m_YComp*source.m_ZComp - m_ZComp*source.m_YComp,
	dest.m_YComp = m_ZComp*source.m_XComp - m_XComp*source.m_ZComp,
	dest.m_ZComp = m_XComp*source.m_YComp - m_YComp*source.m_XComp;
}


float CVector3D::Magnitude()
{
	return (float)sqrt(m_XComp * m_XComp + 
					   m_YComp * m_YComp + 
					   m_ZComp * m_ZComp);
}


void CVector3D::Normalize()
{
	float magnitude = Magnitude();

	if(magnitude > 0.0f)
	{
		m_XComp = m_XComp / magnitude;
		m_YComp = m_YComp / magnitude;
		m_ZComp = m_ZComp / magnitude;
	}
}

//! rotates vector around axis(must be unit vector)
void CVector3D::Rotate(int angle, CVector3D &axis)
{
	// using quaternions (vector and scalar part)

	float s = m_CosTable.GetCos(angle >> 1);
	float h = m_SinTable.GetSin(angle >> 1);

	m_HVec[0].m_XComp = axis.m_XComp * h;
	m_HVec[0].m_YComp = axis.m_YComp * h;
	m_HVec[0].m_ZComp = axis.m_ZComp * h;

	m_HVec[0].CrossProduct(*this,m_HVec[1]);
	m_HVec[0].CrossProduct(m_HVec[1],m_HVec[2]);

	float p_v_dot = DotProduct(m_HVec[0]);

	float newXComp = s * s * m_XComp + m_HVec[0].m_XComp * p_v_dot + 
					 2.0f * s *(m_HVec[1].m_XComp) + m_HVec[2].m_XComp;

	float newYComp = s * s * m_YComp + m_HVec[0].m_YComp * p_v_dot + 
					 2.0f * s *(m_HVec[1].m_YComp) + m_HVec[2].m_YComp;

	float newZComp = s * s * m_ZComp + m_HVec[0].m_ZComp * p_v_dot + 
					 2.0f * s *(m_HVec[1].m_ZComp) + m_HVec[2].m_ZComp;

	m_XComp = newXComp;
	m_YComp = newYComp;
	m_ZComp = newZComp;
}

/*!
	Computes normal by setting one component 0(usually the z-component), 
	exchange the two others and negating one of the two other components(usually 
	the y-component)
	
	if either the x or the y-component is not 0, then the normal vector is in
	the x-y plane, else it is in the x-z plane.
	for example: normal([1,4,5]) = [-4,1,0]
*/
void CVector3D::Normal(CVector3D &normal)
{
	normal.m_XComp = 0.0f;
	normal.m_YComp = 0.0f;
	normal.m_ZComp = 0.0f;

	// to avoid 0-vector: look if at least one component is not 0)
	if((m_XComp != 0.0f) || (m_YComp != 0.0f))
	{
		normal.m_XComp = -m_YComp;
		normal.m_YComp =  m_XComp;
	}	
	else
	{
		normal.m_XComp = -m_ZComp;
		normal.m_YComp =  m_XComp;
	}
}
/*!
	returns an array of normals with length 1 by computing the normal in the x-y plane
	(if there is one, else x-z plane), then rotating this normal around the axis 
	specified by this vector 'num-1' times with angle 360/'num'(resulting in 'num' other 
	normals
*/
void CVector3D::Normals(CVector3D *normals, int num)
{
	CVector3D axis = *this;
	// vector must be unit vector!
	axis.Normalize();

	Normals(axis,normals,num);
}

void CVector3D::SetComp(float xComp, float yComp, float zComp)
{
	m_XComp = xComp;
	m_YComp = yComp;
	m_ZComp = zComp;
}

bool CVector3D::IsNullVector()
{
	return ((m_XComp == 0.0f) && (m_YComp == 0.0f) && (m_ZComp == 0.0f));
}

/*!
	returns an array of normals with length 1 by computing the normal in the x-y plane
	(if there is one, else x-z plane), then rotating this normal around the axis 
	specified by unitVec 'num-1' times with angle 360/'num'(resulting in 'num' other 
	normals
*/
void CVector3D::Normals(CVector3D &unitVec, CVector3D *normals, int num)
{
	unitVec.Normal(normals[0]);
	normals[0].Normalize();
	
	for(int i=1;i<num;i++)
	{
		// Set normal i to normal 0, so you can rotate normal 0.
		normals[i].m_XComp = normals[0].m_XComp;
		normals[i].m_YComp = normals[0].m_YComp;
		normals[i].m_ZComp = normals[0].m_ZComp;

		// Compute angle(rotating 360 deg in "num" steps).
		int angle = (DEF_ANGLE_360 * i) / num;

		// Now rotate.
		normals[i].Rotate(angle,unitVec);
	}
}

void CVector3D::Mult(float *matrix4D)
{
	float xComp = m_XComp;
	float yComp = m_YComp;
	float zComp = m_ZComp;

	m_XComp = xComp*matrix4D[0] + yComp*matrix4D[1] + zComp*matrix4D[2];
	m_YComp = xComp*matrix4D[4] + yComp*matrix4D[5] + zComp*matrix4D[6];
	m_ZComp = xComp*matrix4D[8] + yComp*matrix4D[9] + zComp*matrix4D[10];
}
