// CVectorField.cpp: implementation of the CVectorField class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Flow3D.h"
#include "VectorField.h"
#include "FLOAT.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CVectorField::CVectorField()
{
	m_VField = NULL;
	m_XDim = 0;
	m_YDim = 0;
	m_ZDim = 0;

	m_MaxVectorLength = FLT_MIN;
	m_MinVectorLength = FLT_MAX;
	
	m_MaxPos = CPoint3D(0,0,0);
	m_MinPos = CPoint3D(0,0,0);
	
	m_WinMaxPos = CPoint3D(0,0,0);
	m_WinMinPos = CPoint3D(0,0,0);

	m_SeedMinPos = CPoint3D(0,0,0);
	m_SeedMaxPos = CPoint3D(0,0,0);

	m_MaxAdditional = new float[CSamplePoint::GetAdditionalSize()];
	m_MinAdditional = new float[CSamplePoint::GetAdditionalSize()];

	for(int i=0; i < CSamplePoint::GetAdditionalSize(); i++)
	{
		m_MaxAdditional[i] = FLT_MIN;	
		m_MinAdditional[i] = FLT_MAX;
	}
}

CVectorField::CVectorField(int xDim,int yDim,int zDim)
{
	m_XDim = xDim; 
	m_YDim = yDim; 
	m_ZDim = zDim;
	
	m_VField = NULL;
	m_MaxVectorLength = FLT_MIN;
	m_MinVectorLength = FLT_MAX;

	m_MaxPos = CPoint3D(0,0,0);
	m_MinPos = CPoint3D(0,0,0);
	
	m_WinMaxPos = CPoint3D(0,0,0);
	m_WinMinPos = CPoint3D(0,0,0);

	m_SeedMinPos = CPoint3D(0,0,0);
	m_SeedMaxPos = CPoint3D(0,0,0);

	m_MaxAdditional = NULL;
	m_MinAdditional = NULL;
}

CVectorField::~CVectorField()
{
	if(m_VField!=NULL)
	{
		for(int x=0;x<m_XDim;x++)
		{
			for(int y=0;y<m_YDim;y++)
			{
				delete [] m_VField[x][y];
			}
			delete [] m_VField[x];			
		}
		delete [] m_VField;
		m_VField = NULL;
	}

	if(m_MaxAdditional != NULL)
		delete [] m_MaxAdditional;

	if(m_MinAdditional != NULL)
		delete [] m_MinAdditional;
}

void CVectorField::SetDim(int xDim,int yDim, int zDim)
{
	m_XDim = xDim;
	m_YDim = yDim;
	m_ZDim = zDim;
}

int CVectorField::GetXDim()
{
	return m_XDim;
}

int CVectorField::GetYDim()
{
	return m_YDim;
}

int CVectorField::GetZDim()
{
	return m_ZDim;
}

CSamplePoint * CVectorField::GetVField(int xIndex,int yIndex,int zIndex)
{
	return &m_VField[xIndex][yIndex][zIndex];
}

void CVectorField::SetVField(CSamplePoint &sp,int xIndex,int yIndex,int zIndex)
{
	m_VField[xIndex][yIndex][zIndex] = sp;
}

/*!
	Creates field with analytical formula given by sampler. if sampler is
	NULL, only the memory for the field is allocated.
	CProgressCtrl progress is updated each time the x-index of
	he loop is increased.
	It must be set to NULL if not wanted.
*/
void CVectorField::CreateField(CSampler* sampler, CProgressCtrl *progress)
{
	// Allocate the array and insert samplepoints.
	m_VField = new CSamplePoint**[m_XDim];
	
	for(int x=0;x<m_XDim;x++)
	{
		// Update progresscontrol, if it exists.
		if(progress != NULL)
			progress->StepIt();

		m_VField[x] = new CSamplePoint*[m_YDim];

		for(int y=0;y<m_YDim;y++)
		{
			m_VField[x][y] = new CSamplePoint[m_ZDim];

			if(sampler != NULL)
				for(int z=0;z<m_ZDim;z++)
				{
					sampler->SetPos(CPoint3D((float)x,(float)y,(float)z));
					m_VField[x][y][z] = sampler->ComputeSample();
				
					CSamplePoint *s = &m_VField[x][y][z];

					if(s->GetMagnitude() > m_MaxVectorLength)
						m_MaxVectorLength = s->GetMagnitude();

					if(s->GetMagnitude() < m_MinVectorLength)
						m_MinVectorLength = s->GetMagnitude();
				}
		}
	}

	// Initialising max/min values for this field.
	InitMinMaxPos();
	CalcMinMaxScalars();
}


/*!
	Function uses binary search to get surrounding samplepoints in vector with 8 samplepoints
	for example:
	array[5] = array[B_101] = samplepoint an den Koordinaten [upper-x][upper-y][lower-z]
*/
CSamplePoint ** CVectorField::GetSamplePoints(CPoint3D &pos)
{
	// applying binary search for x-direction
	int lowerX = 0;
	int upperX = m_XDim-1;
	
	do
	{  
		int delim = (lowerX + upperX)/2;//change to shift later
		CSamplePoint *s = &m_VField[delim][0][0];

		if(pos.m_XPos < s->GetXPos())
			upperX = delim;
		else 
			lowerX = delim;
	}
	while(upperX - lowerX > 1);

	// for y-direction
	int lowerY = 0;
	int upperY = m_YDim-1;

	do
	{
		int delim = (lowerY + upperY)/2;
		CSamplePoint *s = &m_VField[0][delim][0];

		if(pos.m_YPos < s->GetYPos())
			upperY = delim;
		
		else 
			lowerY = delim;
	}
	while(upperY - lowerY > 1);
	
	// for z-direction
	int lowerZ = 0;
	int upperZ = m_ZDim-1;

	do
	{
		int delim = (lowerZ + upperZ)/2;
		CSamplePoint *s = &m_VField[0][0][delim];

		if(pos.m_ZPos < s->GetZPos())
			upperZ = delim;
		else 
			lowerZ = delim;
	}
	while(upperZ - lowerZ > 1);

	m_SurSp[0] = &m_VField[lowerX][lowerY][lowerZ];
	m_SurSp[1] = &m_VField[lowerX][lowerY][upperZ];
	m_SurSp[2] = &m_VField[lowerX][upperY][lowerZ];
	m_SurSp[3] = &m_VField[lowerX][upperY][upperZ];
	m_SurSp[4] = &m_VField[upperX][lowerY][lowerZ];
	m_SurSp[5] = &m_VField[upperX][lowerY][upperZ];
	m_SurSp[6] = &m_VField[upperX][upperY][lowerZ];
	m_SurSp[7] = &m_VField[upperX][upperY][upperZ];

	return m_SurSp;
}

void CVectorField::SetMinMaxPos(CPoint3D &minPos, CPoint3D &maxPos)
{
	m_MinPos = minPos;
	m_MaxPos = maxPos;
}

void CVectorField::SetWinMinMaxPos(CPoint3D &minPos, CPoint3D &maxPos)
{
	m_WinMinPos = minPos;
	m_WinMaxPos = maxPos;
}

void CVectorField::SetSeedMinMaxPos(CPoint3D &minPos, CPoint3D &maxPos)
{
	m_SeedMinPos = minPos;
	m_SeedMaxPos = maxPos;
}

float CVectorField::GetMaxVectorLength()
{
	return m_MaxVectorLength;
}

void CVectorField::SetMaxVectorLength(float vectorLength)
{
	m_MaxVectorLength = vectorLength;
}

float CVectorField::GetMinVectorLength()
{
	return m_MinVectorLength;
}

void CVectorField::SetMinVectorLength(float vectorLength)
{
	m_MinVectorLength = vectorLength;
}

float CVectorField::GetMaxAdditional(int index)
{
	return m_MaxAdditional[index];
}

float CVectorField::GetMinAdditional(int index)
{
	return m_MinAdditional[index];
}

void CVectorField::SetMaxAdditional(int index, float value)
{
	m_MaxAdditional[index] = value;
}

void CVectorField::SetMinAdditional(int index, float value)
{
	m_MinAdditional[index] = value;
}

float CVectorField::GetWindowXRange()
{
	return m_WinMaxPos.m_XPos - m_WinMinPos.m_XPos;
}

float CVectorField::GetWindowYRange()
{
	return m_WinMaxPos.m_YPos - m_WinMinPos.m_YPos;
}

float CVectorField::GetWindowZRange()
{
	return m_WinMaxPos.m_ZPos - m_WinMinPos.m_ZPos;
}

float CVectorField::GetXRange()
{
	return m_MaxPos.m_XPos - m_MinPos.m_XPos;
}

float CVectorField::GetYRange()
{
	return m_MaxPos.m_YPos - m_MinPos.m_YPos;
}

float CVectorField::GetZRange()
{
	return m_MaxPos.m_ZPos - m_MinPos.m_ZPos;
}

void CVectorField::InitMinMaxPos()
{
	m_MaxPos = *m_VField[m_XDim-1][m_YDim-1][m_ZDim-1].GetPos();
	m_MinPos = *m_VField[0][0][0].GetPos();
	
	m_WinMaxPos = m_MaxPos;
	m_WinMinPos = m_MinPos;

	m_SeedMaxPos = m_MaxPos;
	m_SeedMinPos = m_MinPos;
}

void CVectorField::CalcMinMaxScalars()
{
	if(m_MaxAdditional != NULL)
		delete m_MaxAdditional;

	if(m_MinAdditional != NULL)
		delete m_MinAdditional;

	m_MaxAdditional = new float[CSamplePoint::GetAdditionalSize()];
	m_MinAdditional = new float[CSamplePoint::GetAdditionalSize()];

	for(int i=0; i < CSamplePoint::GetAdditionalSize(); i++)
	{
		m_MaxAdditional[i] = FLT_MIN;	
		m_MinAdditional[i] = FLT_MAX;
	}

	m_MaxVectorLength = FLT_MIN;
	m_MinVectorLength = FLT_MAX;

	for(int x=0;x<m_XDim;x++)
		for(int y=0;y<m_YDim;y++)
			for(int z=0;z<m_ZDim;z++)
			{
				CSamplePoint *s = &m_VField[x][y][z];

				if(s->GetMagnitude() > m_MaxVectorLength)
					m_MaxVectorLength = s->GetMagnitude();

				if(s->GetMagnitude() < m_MinVectorLength)
					m_MinVectorLength = s->GetMagnitude();

				for(int i=0; i < CSamplePoint::GetAdditionalSize(); i++)
				{
					if(s->GetAdditional(i) > m_MaxAdditional[i])
						m_MaxAdditional[i] = s->GetAdditional(i);
		
					if(s->GetAdditional(i) < m_MinAdditional[i])
						m_MinAdditional[i] = s->GetAdditional(i);
				}
			}
}
