// CVectorField.h: interface for the CVectorField class.
//
//////////////////////////////////////////////////////////////////////
#include "Sampler.h"
#include "Defines.h"

#if !defined(AFX_VECTORFIELD_H__379CDB9A_8EBC_46E2_8CE3_F634ADB1C064__INCLUDED_)
#define AFX_VECTORFIELD_H__379CDB9A_8EBC_46E2_8CE3_F634ADB1C064__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//!  A class representing a vectorfield. 
/*!
	This class consists of samplepoints approximating a 3D-flowfield
	The grid can be rectilinear or cartesian. The class also contains
	a grid with cell size dSep. which is used for the valitation of new 
	samplepoints. This dSep grid stores a list of samplepoint for every
	grid cell.
*/
class CVectorField
{

public:
	void CalcMinMaxScalars();
	//! Allocates memory for the field.
	void InitMinMaxPos();
	//! Returns range of field in x-dimension.
	float GetXRange();
	//! Returns range of field in y-dimension.
	float GetYRange();
	//! Returns range of field in z-dimension.
	float GetZRange();

	//! Returns range of field window in x-dimension.
	float GetWindowXRange();
	//! Returns range of field window in y-dimension.
	float GetWindowYRange();
	//! Returns range of field window in z-dimension.
	float GetWindowZRange();

	//! Sets minimal additional info value on specified index.
	void SetMinAdditional(int index, float value);
	//! Sets minimal additional info value on specified index.
	void SetMaxAdditional(int index, float value);
	//! Returns minimal additional info value on specified index.
	float GetMinAdditional(int index);
	//! Returns maximal additional info value on specified index.
	float GetMaxAdditional(int index);
	//! Sets minimal vector length of field.
	void SetMinVectorLength(float vectorLength);
	//! Returns minimal vector length of field.
	float GetMinVectorLength();
	//! Sets maximal vector length of field.
	void SetMaxVectorLength(float vectorLength);
	//! Returns maximal vector length of field.
	float GetMaxVectorLength();

	//! Sets minimum and maximum positions for window of interest in vectorfield.
	void SetWinMinMaxPos(CPoint3D &minPos, CPoint3D &maxPos);
	//! Sets minimum and maximum positions in vectorfield.
	void SetMinMaxPos(CPoint3D &minPos, CPoint3D &maxPos);
	//! Sets minimum and maximum seedpoint positions in vectorfield.
	void SetSeedMinMaxPos(CPoint3D &minPos, CPoint3D &maxPos);

	//! Returns the 8 samplepoints surrounding the position pos.
	CSamplePoint ** GetSamplePoints(CPoint3D &pos);
		
	//! Creates field with analytical formula given by sampler.
	void CreateField(CSampler *sampler, CProgressCtrl *progress);
	//! Sets dimensions of vectorfield.
	void SetDim(int XDim,int YDim, int ZDim);
	//! Sets samplepoint sp on index xIndex, yIndex, zIndex n samplepoint-grid.
	void SetVField(CSamplePoint &sp, int xIndex, int yIndex, int zIndex);
	//! Returns x-dimension of samplepoints-grid.
	int GetXDim();
	//! Returns y-dimension of samplepoints-grid.
	int GetYDim();
	//! Returns z-dimension of samplepoints-grid.
	int GetZDim();
	
	//! Returns samplepoint on specified index n samplepoint-grid.
	CSamplePoint *GetVField(int xIndex, int yIndex, int zIndex);
	
	//! Maximal position value of the samplepoints.
	CPoint3D m_MaxPos;
	//! Minimal position value of the samplepoints.
	CPoint3D m_MinPos;
	
	//! Maximal position value of the calculation window.
	CPoint3D m_WinMaxPos;
	//! Minimal position value of the calculation window.
	CPoint3D m_WinMinPos;

	//! Maximal position value of a new seedpoint.
	CPoint3D m_SeedMaxPos;
	//! Minimal xosition value of a new seedpoint.
	CPoint3D m_SeedMinPos;
	
	//! Initialises vectorfield with samplepoint-grid = NULL.
	CVectorField();
	//! Initialises vectorfield with dimensions xDim,yDim,zDim.
	CVectorField(int xDim,int yDim,int zDim);

	virtual ~CVectorField();
private:
	CSamplePoint *m_SurSp[8];

	float * m_MinAdditional;
	float * m_MaxAdditional;

	float m_MaxVectorLength;
	float m_MinVectorLength;
	
	//! dimensions of vectorfield
	int m_XDim,m_YDim,m_ZDim;

	//! 3-dimensional field of samplepoints
	CSamplePoint *** m_VField;
	
};

#endif // !defined(AFX_VECTORFIELD_H__379CDB9A_8EBC_46E2_8CE3_F634ADB1C064__INCLUDED_)
